import logging

from dranspose.event import ResultData


class TestReducer:
    def __init__(self, state=None, **kwargs):
        self.publish = {"results": {}, "parameters": {}}

    def process_result(self, result: ResultData, parameters=None):
        logging.info("parameters are %s", parameters)
        self.publish["results"][result.event_number] = result.payload
        self.publish["parameters"][result.event_number] = parameters

    def finish(self, parameters=None):
        print("finished dummy reducer work")



import asyncio
import logging
from asyncio import StreamReader, StreamWriter
from typing import Awaitable, Callable, Coroutine, Optional, Any

import aiohttp

import pytest
import zmq
from pydantic_core import Url

from dranspose.helpers.utils import cancel_and_wait
from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.ingesters.tcp_positioncap import TcpPcapIngester, TcpPcapSettings
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)

from dranspose.worker import Worker, WorkerSettings
from tests.utils import wait_for_controller, wait_for_finish


async def handle_client(reader: StreamReader, writer: StreamWriter) -> None:
    logging.error("ingester connected to pcap")


@pytest.mark.asyncio
async def test_pcapingester(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
) -> None:
    server = await asyncio.start_server(handle_client, "localhost", 8889)
    task = asyncio.create_task(server.serve_forever())

    await reducer(None)
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w1"),
                worker_class="examples.parser.positioncap:PcapWorker",
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )
    await create_ingester(
        TcpPcapIngester(
            settings=TcpPcapSettings(
                ingester_streams=[StreamName("pcap")],
                upstream_url=Url("tcp://localhost:8889"),
                ingester_url=Url("tcp://localhost:10012"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("pcap"), StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        await resp.json()

    context = zmq.asyncio.Context()
    await stream_eiger(context, 9999, ntrig - 1)

    content = await wait_for_finish()

    await cancel_and_wait(task)
    context.destroy()
    print(content)



import logging
from typing import Callable, Optional, Awaitable

import aiohttp
import pytest
from pydantic_core import Url


from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.parameters import ParameterList
from dranspose.protocol import WorkerName, StreamName
from dranspose.worker import Worker, WorkerSettings
from tests.utils import wait_for_controller


@pytest.mark.asyncio
async def test_default_params(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
) -> None:
    await reducer("examples.dummy.reducer:FluorescenceReducer")
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w1"),
                worker_class="examples.dummy.worker:FluorescenceWorker",
            ),
        )
    )

    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        par = await session.get("http://localhost:5000/api/v1/parameters")
        assert par.status == 200
        params = ParameterList.validate_python(await par.json())

        logging.warning("params %s", params)

        resp = await session.get(
            "http://localhost:5000/api/v1/parameter/roi1",
        )
        assert resp.status == 200
        assert await resp.content.read() == b"bla"

        resp = await session.get(
            "http://localhost:5000/api/v1/parameter/file_parameter",
        )
        assert resp.status == 200
        assert await resp.content.read() == b""



from dranspose.ingesters.tcp_positioncap import (  # noqa: F401
    TcpPcapIngester,
    TcpPcapSettings,
)
from dranspose.ingesters.http_sardana import (  # noqa: F401
    HttpSardanaIngester,
    HttpSardanaSettings,
)
from dranspose.ingesters.zmqpull_single import (  # noqa: F401
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.ingesters.zmqpull_eiger_legacy import (  # noqa: F401
    ZmqPullEigerLegacyIngester,
    ZmqPullEigerLegacySettings,
)
from dranspose.ingesters.zmqsub_pcap import (  # noqa: F401
    ZmqSubPCAPIngester,
    ZmqSubPCAPSettings,
)
from dranspose.ingesters.zmqsub_contrast import (  # noqa: F401
    ZmqSubContrastIngester,
    ZmqSubContrastSettings,
)
from dranspose.ingesters.zmqsub_xspress3 import (  # noqa: F401
    ZmqSubXspressIngester,
    ZmqSubXspressSettings,
)
from dranspose.ingesters.zmqsub_albaem import (  # noqa: F401
    ZmqSubAlbaemIngester,
    ZmqSubAlbaemSettings,
)
from dranspose.ingesters.zmqsub_lecroy import (  # noqa: F401
    ZmqSubLecroyIngester,
    ZmqSubLecroySettings,
)



import logging
import time

from dranspose.event import EventData
from dranspose.parameters import FloatParameter

logger = logging.getLogger(__name__)


class SlowWorker:
    def __init__(self, **kwargs):
        pass

    @staticmethod
    def describe_parameters():
        params = [
            FloatParameter(name="sleep_time"),
        ]
        return params

    def process_event(self, event: EventData, parameters=None, *args, **kwargs):
        logger.debug("using parameters %s", parameters)
        if "sleep_time" in parameters:
            time.sleep(parameters["sleep_time"].value)
        else:
            time.sleep(0.28)
        return "not empty"



import logging
from datetime import datetime, timezone

from dranspose.event import EventData
from dranspose.middlewares import stream1

logger = logging.getLogger(__name__)


class TimingWorker:
    def __init__(self, **kwargs):
        pass

    def process_event(
        self, event: EventData, parameters=None, tick=False, *args, **kwargs
    ):
        logger.debug("using parameters %s", parameters)
        logger.info("tick %s", tick)
        times = {}
        for stream in event.streams:
            if stream in ["large", "fast"]:
                dat = stream1.parse(event.streams[stream])
                if hasattr(dat, "timestamps"):
                    ts = dat.timestamps
                    now = datetime.now(timezone.utc)
                    oldest = min(map(lambda x: datetime.fromisoformat(x), ts.values()))
                    delta = now - oldest
                    logger.debug("delta is %s", delta.total_seconds())
                    times[stream] = (oldest, delta.total_seconds())
        return times

    def get_tick_interval(self, parameters=None):
        return 1



import json
from types import UnionType
from typing import Any

import zmq

from dranspose.data.eiger_legacy import (
    EigerLegacyHeader,
    EigerLegacyPacket,
    EigerLegacyImage,
)
from dranspose.event import StreamData


def _get_json(frame: zmq.Frame | bytes) -> Any:
    if isinstance(frame, zmq.Frame):
        val = json.loads(frame.bytes)
    elif isinstance(frame, bytes):
        val = json.loads(frame)
    else:
        raise Exception("invalid StreamData")
    return val


def parse(data: StreamData) -> UnionType:
    """
    Parses a eiger legacy packet, which returns a start,image or end message

    Arguments:
        data: a frame comming from a eiger legacy source

    Returns:
        a EigerLegacyPacket
    """
    assert data.typ == "EIGER_LEGACY"
    packet = EigerLegacyPacket.validate_python(_get_json(data.frames[0]))
    if isinstance(packet, EigerLegacyHeader):
        assert data.length == 9
        packet.info = _get_json(data.frames[1])
        packet.appendix = _get_json(data.frames[8])

    if isinstance(packet, EigerLegacyImage):
        assert data.length == 5
        info = _get_json(data.frames[1])
        buffer = data.frames[2]
        if isinstance(buffer, zmq.Frame):
            buffer = buffer.bytes
        info["buffer"] = buffer
        packet.data = info
        packet.config = _get_json(data.frames[3])

    return packet



import asyncio
import json
import logging
import pickle
import random
import socket
import string
import time
import traceback
from asyncio import Future, Task
from typing import Optional, Any

import zmq.asyncio

from pydantic import UUID4, BaseModel, ConfigDict, Field

from dranspose.helpers import utils
import redis.exceptions as rexceptions

from dranspose.distributed import DistributedService, DistributedSettings
from dranspose.event import InternalWorkerMessage, EventData, ResultData
from dranspose.helpers.utils import done_callback, cancel_and_wait
from dranspose.protocol import (
    WorkerState,
    RedisKeys,
    IngesterState,
    WorkerUpdate,
    DistributedStateEnum,
    WorkerName,
    IngesterName,
    StreamName,
    ReducerState,
    WorkerTimes,
    GENERIC_WORKER,
    WorkerTag,
    WorkAssignmentList,
)


class RedisException(Exception):
    pass


class ConnectedIngester(BaseModel):
    "An ingester becomes a connected ingester once the worker opened a socket to it"
    model_config = ConfigDict(arbitrary_types_allowed=True)
    socket: zmq.asyncio.Socket
    config: IngesterState
    pinged_since: float = Field(default_factory=time.time)


def random_worker_name() -> WorkerName:
    "without a given worker name, generate a random one"
    randid = "".join([random.choice(string.ascii_letters) for _ in range(10)])
    name = "Worker-{}-{}".format(socket.gethostname(), randid)
    return WorkerName(name)


class WorkerSettings(DistributedSettings):
    worker_class: Optional[str] = None
    worker_name: WorkerName = Field(default_factory=random_worker_name)
    worker_tags: set[WorkerTag] = {GENERIC_WORKER}


class Worker(DistributedService):
    def __init__(self, settings: Optional[WorkerSettings] = None):
        self._worker_settings = settings
        if self._worker_settings is None:
            self._worker_settings = WorkerSettings()

        state = WorkerState(
            name=self._worker_settings.worker_name,
            tags=self._worker_settings.worker_tags,
        )
        super().__init__(state, self._worker_settings)
        self._logger.info("created worker with state %s", state)
        self.state: WorkerState
        self.ctx = zmq.asyncio.Context()

        self._ingesters: dict[IngesterName, ConnectedIngester] = {}
        self._stream_map: dict[StreamName, zmq._future._AsyncSocket] = {}
        self.poll_task: Optional[Future[list[int]]] = None

        self._reducer_service_uuid: Optional[UUID4] = None
        self.out_socket: Optional[zmq._future._AsyncSocket] = None

        self.assignment_queue: asyncio.Queue[
            set[zmq._future._AsyncSocket]
        ] = asyncio.Queue()
        self.dequeue_task: Optional[Task[set[zmq._future._AsyncSocket]]] = None

        self.param_descriptions = []
        self.custom = None
        self.custom_context: dict[Any, Any] = {}
        self.worker = None
        if self._worker_settings.worker_class:
            try:
                self.custom = utils.import_class(self._worker_settings.worker_class)
                self._logger.info("custom worker class %s", self.custom)

                try:
                    self.param_descriptions = self.custom.describe_parameters()  # type: ignore[attr-defined]
                except AttributeError:
                    self._logger.info(
                        "custom worker class has no describe_parameters staticmethod"
                    )
                except Exception as e:
                    self._logger.error(
                        "custom worker parameter descripition is broken: %s",
                        e.__repr__(),
                    )

            except Exception as e:
                self._logger.error(
                    "no custom worker class loaded, discarding events, err %s\n%s",
                    e.__repr__(),
                    traceback.format_exc(),
                )

    async def run(self) -> None:
        self.manage_ingester_task = asyncio.create_task(self.manage_ingesters())
        self.manage_ingester_task.add_done_callback(done_callback)
        self.manage_receiver_task = asyncio.create_task(self.manage_receiver())
        self.manage_receiver_task.add_done_callback(done_callback)
        self.assignment_queue = asyncio.Queue()
        self.work_task = asyncio.create_task(self.work())
        self.work_task.add_done_callback(done_callback)
        self.assign_task = asyncio.create_task(self.manage_assignments())
        self.assign_task.add_done_callback(done_callback)
        self.metrics_task = asyncio.create_task(self.update_metrics())
        self.metrics_task.add_done_callback(done_callback)
        await self.register()

    async def notify_worker_ready(self) -> None:
        "send an update to the controller that the worker got the new trigger map and is ready to receive the first event"
        await self.redis.xadd(
            RedisKeys.ready(self.state.mapping_uuid),
            {
                "data": WorkerUpdate(
                    state=DistributedStateEnum.READY,
                    worker=self.state.name,
                ).model_dump_json()
            },
        )

        self._logger.info("registered ready message")

    async def manage_assignments(self) -> None:
        """
        This coroutine listens to the redis stream for the current mapping_uuid.
        Once it there is a new batch of assignments, it checks for each,
        if this worker is involved and if yes, from which ingesters it should receive.
        It places the list of ingesters in the assignment_queue.
        """
        sub = RedisKeys.assigned(self.state.mapping_uuid)
        lastev = 0
        while True:
            try:
                assignments = await self.redis.xread({sub: lastev}, block=1000, count=1)
            except rexceptions.ConnectionError:
                break
            if sub not in assignments:
                continue
            assignments = assignments[sub][0][0]
            self._logger.debug("got assignments %s", assignments)
            self._logger.debug("stream map %s", self._stream_map)
            work_assignment_list = WorkAssignmentList.validate_json(
                assignments[1]["data"]
            )
            for work_assignment in work_assignment_list:
                ingesterset = set()
                for stream, workers in work_assignment.assignments.items():
                    if self.state.name in workers:
                        try:
                            ingesterset.add(self._stream_map[stream])
                        except KeyError:
                            self._logger.error(
                                "ingester for stream %s not connected, available: %s",
                                stream,
                                self._ingesters,
                            )
                self._logger.debug("receive from ingesters %s", ingesterset)
                if len(ingesterset) > 0:
                    await self.assignment_queue.put(ingesterset)
            lastev = assignments[0]

    async def poll_internals(
        self, ingesterset: set[zmq._future._AsyncSocket]
    ) -> list[int]:
        """
        A task to simultaneously check if there is a message available from all required ingesters
        We do not yet receive from the zmq sockets as it is harder to cancel a receive call.
        """
        self._logger.debug("poll internal sockets %s", ingesterset)
        poll_tasks = [sock.poll() for sock in ingesterset]
        self._logger.debug("await poll tasks %s", poll_tasks)
        self.poll_task = asyncio.gather(*poll_tasks)
        done = await self.poll_task
        self._logger.debug("data is available done: %s", done)
        return done

    async def build_event(
        self, ingesterset: set[zmq._future._AsyncSocket]
    ) -> EventData:
        "All relevant ingester sockets have a message, receive it and assemble an EventData object"
        msgs = []
        for sock in ingesterset:
            res = await sock.recv_multipart(copy=False)
            prelim = json.loads(res[0].bytes)
            pos = 1
            for stream, data in prelim["streams"].items():
                data["frames"] = res[pos : pos + data["length"]]
                pos += data["length"]
            msg = InternalWorkerMessage.model_validate(prelim)
            msgs.append(msg)

        return EventData.from_internals(msgs)

    async def work(self) -> None:
        self._logger.info("started work task")

        self.worker = None
        if self.custom:
            try:
                self.worker = self.custom(
                    parameters=self.parameters,
                    context=self.custom_context,
                    state=self.state,
                )
            except Exception as e:
                self._logger.error(
                    "Failed to instantiate custom worker: %s", e.__repr__()
                )

        await self.notify_worker_ready()
        try:
            proced = 0
            completed = []
            has_result = []
            accum_times = WorkerTimes(no_events=0)
            accum_start = time.time()
            tick_wait_until = time.time() - 1
            while True:
                perf_start = time.perf_counter()

                self.dequeue_task = None
                self.dequeue_task = asyncio.create_task(self.assignment_queue.get())
                ingesterset = await self.dequeue_task
                perf_got_assignments = time.perf_counter()
                done = await self.poll_internals(ingesterset)
                if set(done) != {zmq.POLLIN}:
                    self._logger.warning("not all sockets are pollIN %s", done)
                    continue

                perf_got_work = time.perf_counter()

                event = await self.build_event(ingesterset)

                perf_assembled_event = time.perf_counter()
                self._logger.debug("received work %s", event)
                result = None
                if self.worker:
                    tick = False
                    # we internally cache when the redis will expire to reduce hitting redis on every event
                    if tick_wait_until - time.time() < 0:
                        if hasattr(self.worker, "get_tick_interval"):
                            wait_ms = int(
                                self.worker.get_tick_interval(self.parameters) * 1000
                            )
                            dist_clock = await self.redis.set(
                                RedisKeys.clock(self.state.mapping_uuid),
                                "🕙",
                                px=wait_ms,
                                nx=True,
                            )
                            if dist_clock is True:
                                tick = True
                            else:
                                expire_ms = await self.redis.pttl(
                                    RedisKeys.clock(self.state.mapping_uuid)
                                )
                                tick_wait_until = time.time() + (expire_ms / 1000)
                    try:
                        loop = asyncio.get_event_loop()
                        result = await loop.run_in_executor(
                            None,
                            self.worker.process_event,
                            event,
                            self.parameters,
                            tick,
                        )
                    except Exception as e:
                        self._logger.error(
                            "custom worker failed: %s\n%s",
                            e.__repr__(),
                            traceback.format_exc(),
                        )
                perf_custom_code = time.perf_counter()
                self._logger.debug("got result %s", result)
                if result is not None:
                    rd = ResultData(
                        event_number=event.event_number,
                        worker=self.state.name,
                        payload=result,
                        parameters_hash=self.state.parameters_hash,
                    )
                    if self.out_socket:
                        try:
                            header = rd.model_dump_json(exclude={"payload"}).encode(
                                "utf8"
                            )
                            body = pickle.dumps(rd.payload)
                            self._logger.debug(
                                "send result to reducer with header %s, len-payload %d",
                                header,
                                len(body),
                            )
                            await self.out_socket.send_multipart([header, body])
                        except Exception as e:
                            self._logger.error(
                                "could not send out result %s", e.__repr__()
                            )
                perf_sent_result = time.perf_counter()
                proced += 1
                self.state.processed_events += 1
                if proced % 500 == 0:
                    self._logger.info("processed %d events", proced)
                completed.append(event.event_number)
                has_result.append(result is not None)
                times = WorkerTimes.from_timestamps(
                    perf_start,
                    perf_got_assignments,
                    perf_got_work,
                    perf_assembled_event,
                    perf_custom_code,
                    perf_sent_result,
                )
                accum_times += times
                # the worker only sends an update to the controller if it is idle or a second has passed
                # this batching is necessary for high frequency event streams
                if self.assignment_queue.empty() or time.time() - accum_start > 1:
                    wu = WorkerUpdate(
                        state=DistributedStateEnum.IDLE,
                        completed=completed,
                        worker=self.state.name,
                        has_result=has_result,
                        processing_times=accum_times,
                    )
                    self._logger.debug("all work done, notify controller with %s", wu)
                    await self.redis.xadd(
                        RedisKeys.ready(self.state.mapping_uuid),
                        {"data": wu.model_dump_json()},
                    )
                    completed = []
                    has_result = []
                    accum_times = WorkerTimes(no_events=0)
                    accum_start = time.time()
        except asyncio.exceptions.CancelledError:
            pass
        self._logger.info("work thread finished")

    async def finish_work(self) -> None:
        self._logger.info("finishing work")
        if self.worker:
            if hasattr(self.worker, "finish"):
                try:
                    loop = asyncio.get_event_loop()
                    await loop.run_in_executor(
                        None, self.worker.finish, self.parameters
                    )
                except Exception as e:
                    self._logger.error(
                        "custom worker finish failed: %s\n%s",
                        e.__repr__(),
                        traceback.format_exc(),
                    )
        await self.redis.xadd(
            RedisKeys.ready(self.state.mapping_uuid),
            {
                "data": WorkerUpdate(
                    state=DistributedStateEnum.FINISHED,
                    worker=self.state.name,
                ).model_dump_json()
            },
        )

    async def restart_work(
        self, new_uuid: UUID4, active_streams: list[StreamName]
    ) -> None:
        self._logger.info("resetting config %s", new_uuid)
        if self.poll_task:
            await cancel_and_wait(self.poll_task)
            self.poll_task = None
            self._logger.debug("cancelled poll task")
        await cancel_and_wait(self.work_task)
        self._logger.info("clean up in sockets")
        await cancel_and_wait(self.assign_task)
        for iname, ing in self._ingesters.items():
            while True:
                res = await ing.socket.poll(timeout=0.001)
                if res == zmq.POLLIN:
                    await ing.socket.recv_multipart(copy=False)
                    self._logger.debug("discarded internal message from %s", iname)
                else:
                    break

        self.assignment_queue = asyncio.Queue()
        self.state.mapping_uuid = new_uuid
        self.work_task = asyncio.create_task(self.work())
        self.work_task.add_done_callback(done_callback)
        self.assign_task = asyncio.create_task(self.manage_assignments())
        self.assign_task.add_done_callback(done_callback)

    async def manage_receiver(self) -> None:
        "Periodically check that the push socket to the receiver is connected to the correct url, which the reducer publishes in redis"
        while True:
            config = await self.redis.get(RedisKeys.config("reducer"))
            if config is None:
                self._logger.warning("cannot get reducer configuration")
                await asyncio.sleep(1)
                continue
            cfg = ReducerState.model_validate_json(config)
            if cfg.service_uuid != self._reducer_service_uuid:
                # connect to a new reducer
                if self.out_socket is not None:
                    self.out_socket.close()
                self.out_socket = self.ctx.socket(zmq.PUSH)
                self.out_socket.connect(str(cfg.url))
                self._reducer_service_uuid = cfg.service_uuid
                self._logger.info("connected out_socket to reducer at %s", cfg.url)
            await asyncio.sleep(10)

    async def manage_ingesters(self) -> None:
        """
        This function is supposed to run in the background and make sure
        that the worker has a DEALER socket to every ingester.
        It periodically pings the ingesters and checks that the pings arrive
        by fetching the state of the ingesters from redis.
        """
        while True:
            configs = await self.redis.keys(RedisKeys.config("ingester"))
            processed = []
            for key in configs:
                raw_config = await self.redis.get(key)
                if raw_config is None:
                    logging.warning(
                        "ingester config key %s disappeard while updating", key
                    )
                    continue
                cfg = IngesterState.model_validate_json(raw_config)
                iname = cfg.name
                processed.append(iname)
                if iname in self._ingesters:
                    pinged_for_long = (
                        time.time() - self._ingesters[iname].pinged_since > 10
                    )
                    reached_ingester = (
                        self.state.service_uuid in cfg.connected_workers.keys()
                    )

                    if self._ingesters[iname].config.service_uuid != cfg.service_uuid:
                        self._logger.warning(
                            "service_uuid of ingester changed from %s to %s, disconnecting",
                            self._ingesters[iname].config.service_uuid,
                            cfg.service_uuid,
                        )
                        self._ingesters[iname].socket.close()
                        del self._ingesters[iname]

                    elif pinged_for_long and not reached_ingester:
                        self._logger.warning(
                            "we send pings, but don't reach ingester %s, disconnecting",
                            iname,
                        )
                        self._ingesters[iname].socket.close()
                        del self._ingesters[iname]

                if iname not in self._ingesters:
                    self._logger.info("adding new ingester %s", iname)
                    try:
                        sock = self.ctx.socket(zmq.DEALER)
                        sock.setsockopt(zmq.IDENTITY, self.state.name.encode("ascii"))
                        sock.connect(str(cfg.url))
                        await sock.send(self.state.service_uuid.bytes)
                        self._ingesters[iname] = ConnectedIngester(
                            config=cfg, socket=sock
                        )
                    except zmq.error.ZMQError:
                        logging.error("cannot open dealer socket to ingester %s", iname)

                await self._ingesters[iname].socket.send(self.state.service_uuid.bytes)
                self._logger.debug("pinged %s", iname)
            for iname in set(self._ingesters.keys()) - set(processed):
                self._logger.info("removing stale ingester %s", iname)
                self._ingesters[iname].socket.close()
                del self._ingesters[iname]
            self._stream_map = {
                s: conn_ing.socket
                for ing, conn_ing in self._ingesters.items()
                for s in conn_ing.config.streams
            }
            new_ingesters = [a.config for a in self._ingesters.values()]
            if self.state.ingesters != new_ingesters:
                self.state.ingesters = new_ingesters
                self._logger.info(
                    "changed ingester config, fast publish %s", new_ingesters
                )
                await self.publish_config()

            await asyncio.sleep(2)

    async def close(self) -> None:
        if self.worker:
            if hasattr(self.worker, "close"):
                try:
                    loop = asyncio.get_event_loop()
                    await loop.run_in_executor(
                        None, self.worker.close, self.custom_context
                    )
                except Exception as e:
                    self._logger.error(
                        "custom worker failed to close: %s\n%s",
                        e.__repr__(),
                        traceback.format_exc(),
                    )
        await cancel_and_wait(self.manage_ingester_task)
        await cancel_and_wait(self.manage_receiver_task)
        await cancel_and_wait(self.metrics_task)
        await cancel_and_wait(self.assign_task)
        if self.dequeue_task is not None:
            await cancel_and_wait(self.dequeue_task)
        await self.redis.delete(RedisKeys.config("worker", self.state.name))
        await super().close()
        self.ctx.destroy()
        self._logger.info("worker closed")



import asyncio
import gzip
import json
import logging
import pickle
import shutil
import threading
from typing import Awaitable, Callable, Any, Coroutine, Optional, IO

import aiohttp
import cbor2
import h5pyd
import numpy as np

import pytest
import zmq.asyncio
import zmq
from pydantic_core import Url

from dranspose.event import EventData
from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)

from dranspose.replay import replay
from dranspose.worker import Worker
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.skipif("config.getoption('rust')", reason="rust does not support dumping")
@pytest.mark.asyncio
async def test_replay(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
    stream_orca: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
    stream_small: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
    tmp_path: Any,
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await create_worker(WorkerName("w2"))
    await create_worker(WorkerName("w3"))

    p_eiger = tmp_path / "eiger_dump.cbors"
    print(p_eiger, type(p_eiger))

    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
                dump_path=p_eiger,
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("orca")],
                upstream_url=Url("tcp://localhost:9998"),
                ingester_url=Url("tcp://localhost:10011"),
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("alba")],
                upstream_url=Url("tcp://localhost:9997"),
                ingester_url=Url("tcp://localhost:10012"),
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("slow")],
                upstream_url=Url("tcp://localhost:9996"),
                ingester_url=Url("tcp://localhost:10013"),
            ),
        )
    )

    await wait_for_controller(
        streams={
            StreamName("eiger"),
            StreamName("orca"),
            StreamName("alba"),
            StreamName("slow"),
        }
    )
    async with aiohttp.ClientSession() as session:
        p_prefix = tmp_path / "dump_"
        logging.info("prefix is %s encoded: %s", p_prefix, str(p_prefix).encode("utf8"))

        resp = await session.post(
            "http://localhost:5000/api/v1/parameter/dump_prefix",
            data=str(p_prefix).encode("utf8"),
        )
        assert resp.status == 200

        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
                "orca": [
                    [
                        VirtualWorker(
                            constraint=VirtualConstraint(2 * i + 1)
                        ).model_dump(mode="json")
                    ]
                    for i in range(1, ntrig)
                ],
                "alba": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        ),
                        VirtualWorker(
                            constraint=VirtualConstraint(2 * i + 1)
                        ).model_dump(mode="json"),
                    ]
                    for i in range(1, ntrig)
                ],
                "slow": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        ),
                        VirtualWorker(
                            constraint=VirtualConstraint(2 * i + 1)
                        ).model_dump(mode="json"),
                    ]
                    if i % 4 == 0
                    else None
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        uuid = await resp.json()

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1))
    asyncio.create_task(stream_orca(context, 9998, ntrig - 1))
    asyncio.create_task(stream_small(context, 9997, ntrig - 1))
    asyncio.create_task(stream_small(context, 9996, ntrig // 4))

    await wait_for_finish()

    # read dump

    par_file = tmp_path / "parameters.json"

    with open(par_file, "w") as f:
        json.dump([{"name": "roi1", "data": "[10,20]"}], f)

    replay(
        "examples.dummy.worker:FluorescenceWorker",
        "examples.dummy.reducer:FluorescenceReducer",
        [p_eiger, f"{p_prefix}orca-ingester-{uuid}.cbors"],
        None,
        par_file,
    )

    with open(f"{p_prefix}orca-ingester-{uuid}.cbors", "rb") as f_in:
        with gzip.open(f"{p_prefix}orca-ingester-{uuid}.cbors.gz", "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)

    stop_event = threading.Event()
    done_event = threading.Event()

    thread = threading.Thread(
        target=replay,
        args=(
            "examples.test.worker:TestWorker",
            "examples.test.reducer:TestReducer",
            [p_eiger, f"{p_prefix}orca-ingester-{uuid}.cbors.gz"],
            None,
            par_file,
        ),
        kwargs={"port": 5010, "stop_event": stop_event, "done_event": done_event},
    )
    thread.start()

    logging.info("keep webserver alive")
    done_event.wait()
    logging.info("replay done")

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5010/api/v1/result/")
        content = await st.content.read()
        result = pickle.loads(content)[0]
        assert list(result["results"].keys()) == list(range(ntrig + 1))
        logging.warning("params in reducer is %s", result["results"][5][0])
        ev5: EventData = result["results"][5][0]
        assert ev5.event_number == 5
        assert ev5.streams[StreamName("eiger")].typ == "STINS"
        eiger_frame_header = ev5.streams[StreamName("eiger")].frames[0]
        eiger_frame_payload = ev5.streams[StreamName("eiger")].frames[1]
        assert isinstance(eiger_frame_header, bytes)
        assert isinstance(eiger_frame_payload, bytes)
        assert eiger_frame_header.startswith(
            b'{"htype": "image", "frame": 4, "shape": [1475, 831], "type": "uint16", "compression": "none", "msg_number": 5, "timestamps": {"dummy": "'
        )
        assert len(eiger_frame_payload) > 4500
        assert ev5.streams[StreamName("orca")].typ == "STINS"
        orca_frame_header = ev5.streams[StreamName("orca")].frames[0]
        orca_frame_payload = ev5.streams[StreamName("orca")].frames[1]
        assert isinstance(orca_frame_header, bytes)
        assert isinstance(orca_frame_payload, bytes)
        assert orca_frame_header.startswith(
            b'{"htype": "image", "frame": 4, "shape": [2000, 4000], "type": "uint16", "compression": "none", "msg_number": 5}'
        )
        assert len(orca_frame_payload) > 4500
    logging.info("shut down server")
    stop_event.set()

    thread.join()

    bin_file = tmp_path / "binparams.pkl"

    fb: IO[bytes]
    with open(bin_file, "wb") as fb:
        arr = np.ones((10, 10))
        pickle.dump(
            [
                {"name": "roi1", "data": "[10,20]"},
                {"name": "file_parameter_file", "data": pickle.dumps(arr)},
            ],
            fb,
        )

    stop_event = threading.Event()
    done_event = threading.Event()

    thread = threading.Thread(
        target=replay,
        args=(
            "examples.dummy.worker:FluorescenceWorker",
            "examples.dummy.reducer:FluorescenceReducer",
            [p_eiger, f"{p_prefix}orca-ingester-{uuid}.cbors"],
            None,
            bin_file,
        ),
        kwargs={"port": 5010, "stop_event": stop_event, "done_event": done_event},
    )
    thread.start()

    logging.info("keep webserver alive")
    done_event.wait()
    logging.info("replay done")

    def work_first() -> None:
        f = h5pyd.File("http://localhost:5010/", "r")
        logging.info("file %s", list(f.keys()))
        logging.info("map %s", list(f["map"].keys()))
        assert list(f.keys()) == ["map"]

    loop = asyncio.get_event_loop()
    await loop.run_in_executor(None, work_first)

    logging.info("shut down server")
    stop_event.set()

    thread.join()

    cbor_file = tmp_path / "binparams.cbor"

    with open(cbor_file, mode="wb") as fb:
        arr = np.ones((10, 10))
        cbor2.dump(
            [
                {"name": "roi1", "data": "[10,20]"},
                {"name": "file_parameter_file", "data": pickle.dumps(arr)},
            ],
            fb,
        )

    stop_event = threading.Event()
    done_event = threading.Event()

    thread = threading.Thread(
        target=replay,
        args=(
            "examples.dummy.worker:FluorescenceWorker",
            "examples.dummy.reducer:FluorescenceReducer",
            [p_eiger, f"{p_prefix}orca-ingester-{uuid}.cbors"],
            None,
            cbor_file,
        ),
        kwargs={"port": 5010, "stop_event": stop_event, "done_event": done_event},
    )
    thread.start()

    logging.info("keep webserver alive")
    done_event.wait()
    logging.info("replay done")

    def work_second() -> None:
        f = h5pyd.File("http://localhost:5010/", "r")
        logging.info("file %s", list(f.keys()))
        logging.info("map %s", list(f["map"].keys()))
        assert list(f.keys()) == ["map"]

    loop = asyncio.get_event_loop()
    await loop.run_in_executor(None, work_second)

    logging.info("shut down server")
    stop_event.set()

    thread.join()

    context.destroy()

    print(content)



import os
import pickle
from pathlib import PosixPath

import asyncio
from typing import Awaitable, Callable, Any, Coroutine, Optional
import zmq.asyncio

import aiohttp
import pytest
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqsub_contrast import (
    ZmqSubContrastIngester,
    ZmqSubContrastSettings,
)
from dranspose.ingesters.zmqsub_xspress3 import (
    ZmqSubXspressIngester,
    ZmqSubXspressSettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)
from dranspose.worker import Worker, WorkerSettings
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_multiple_scans(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_pkls: Callable[
        [zmq.Context[Any], int, os.PathLike[Any] | str, float, int],
        Coroutine[Any, Any, None],
    ],
) -> None:
    await reducer("examples.dummy.reducer:FluorescenceReducer")
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w1"),
                worker_class="examples.dummy.worker:FluorescenceWorker",
            ),
        )
    )
    await create_ingester(
        ZmqSubContrastIngester(
            settings=ZmqSubContrastSettings(
                ingester_streams=[StreamName("contrast")],
                upstream_url=Url("tcp://localhost:5556"),
                ingester_url=Url("tcp://localhost:10000"),
            ),
        )
    )
    await create_ingester(
        ZmqSubXspressIngester(
            settings=ZmqSubXspressSettings(
                ingester_streams=[StreamName("xspress3")],
                upstream_url=Url("tcp://localhost:9999"),
                ingester_url=Url("tcp://localhost:10001"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("contrast"), StreamName("xspress3")})
    async with aiohttp.ClientSession() as session:
        resp = await session.post(
            "http://localhost:5000/api/v1/parameter/roi1",
            json=[0, 10],
        )
        assert resp.status == 200
        await resp.json()

        ntrig = 20
        mp = {
            "contrast": [
                [VirtualWorker(constraint=VirtualConstraint(i)).model_dump(mode="json")]
                for i in range(ntrig)
            ],
            "xspress3": [
                [VirtualWorker(constraint=VirtualConstraint(i)).model_dump(mode="json")]
                for i in range(ntrig)
            ],
        }
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json=mp,
        )
        assert resp.status == 200
        await resp.json()

    context = zmq.asyncio.Context()

    asyncio.create_task(
        stream_pkls(
            context,
            9999,
            PosixPath("tests/data/xspress3mini-dump20.pkls"),
            0.001,
            zmq.PUB,
        )
    )
    asyncio.create_task(
        stream_pkls(
            context, 5556, PosixPath("tests/data/contrast-dump.pkls"), 0.001, zmq.PUB
        )
    )

    await wait_for_finish()
    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5001/api/v1/result/pickle")
        content = await st.content.read()
        result = pickle.loads(content)
        assert result == {
            "map": {
                (-2.004051863, -2.002903037): {"roi1": 81},
                (-2.001404022, -0.9992748245): {"roi1": 76},
                (-2.000941193, 0.001238407776): {"roi1": 67},
                (-2.000656853, 1.001692484): {"roi1": 49},
                (-2.000562814, 2.0019763): {"roi1": 69},
                (-0.6658021229, -2.002587864): {"roi1": 97},
                (-0.6664982812, -0.9992643584): {"roi1": 75},
                (-0.6666898134, 0.001223421204): {"roi1": 58},
                (-0.6666922578, 1.001631271): {"roi1": 81},
                (-0.6666616732, 2.001930725): {"roi1": 77},
                (0.6676852429, -2.002412156): {"roi1": 91},
                (0.6666162969, -0.9992055423): {"roi1": 81},
                (0.6664269385, 0.001214749016): {"roi1": 76},
                (0.6663550388, 1.00170782): {"roi1": 67},
                (0.6663102034, 2.002018839): {"roi1": 49},
                (2.001407847, -2.002363463): {"roi1": 69},
                (2.000183463, -0.9992309104): {"roi1": 97},
                (1.999963432, 0.001258902382): {"roi1": 75},
                (1.999884878, 1.001690624): {"roi1": 58},
                (1.999828501, 2.001959397): {"roi1": 81},
            },
            "version": {
                "commit_hash": "1a474d1455f0d6c6113a22179b8fff1e98325916",
                "branch_name": "main",
                "timestamp": "2024-04-25T08:07:59+02:00",
                "repository_url": "https://gitlab.maxiv.lu.se/DanMAX/pipeline/drp-tomo-buffer",
            },
        }
        assert len(result["map"].keys()) == ntrig

    for _ in range(2):
        await stream_pkls(
            context,
            9999,
            PosixPath("tests/data/xspress3mini-dump20.pkls"),
            0.001,
            zmq.PUB,
        )

    # second scan
    async with aiohttp.ClientSession() as session:
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json=mp,
        )
        assert resp.status == 200
        await resp.json()

    await stream_pkls(
        context,
        9999,
        PosixPath("tests/data/xspress3mini-dump20.pkls"),
        0.001,
        zmq.PUB,
        begin=30,  # type: ignore[call-arg]
    )

    asyncio.create_task(
        stream_pkls(
            context,
            9999,
            PosixPath("tests/data/xspress3mini-dump20.pkls"),
            0.001,
            zmq.PUB,
        )
    )
    asyncio.create_task(
        stream_pkls(
            context, 5556, PosixPath("tests/data/contrast-dump.pkls"), 0.001, zmq.PUB
        )
    )

    await wait_for_finish()
    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5001/api/v1/result/pickle")
        content = await st.content.read()
        result = pickle.loads(content)
        print("content", result)
        assert len(result["map"].keys()) == ntrig

    context.destroy()



import asyncio
import pickle
from typing import Awaitable, Callable, Any, Coroutine, Optional

import aiohttp

import pytest
import zmq.asyncio
import zmq
from pydantic_core import Url

from dranspose.data.stream1 import Stream1Data
from dranspose.event import EventData
from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.middlewares.stream1 import parse
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
    WorkerTag,
)

from dranspose.worker import Worker
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_outside(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    debug_worker: Callable[[Optional[str], Optional[list[str]]], Awaitable[None]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
    time_beacon: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await debug_worker("debugworker", ["debug"])
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )

    context = zmq.asyncio.Context()

    await wait_for_controller(streams={StreamName("eiger")}, workers={WorkerName("w1")})
    async with aiohttp.ClientSession() as session:
        await time_beacon(context, 9999, 5, flags=zmq.NOBLOCK)  # type: ignore[call-arg]
        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(
                            constraint=VirtualConstraint(2 * i),
                            tags={WorkerTag("debug")},
                        ).model_dump(mode="json")
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1))

    await wait_for_finish()

    context.destroy()

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5002/api/v1/last_events?number=20")
        content = await st.content.read()
        res: list[EventData] = pickle.loads(content)
        got_frames: list[int] = []
        for ev in res:
            print("got event", ev.event_number, ev.streams.keys())
            pkt = parse(ev.streams[StreamName("eiger")])
            print(pkt)
            if isinstance(pkt, Stream1Data):
                got_frames.append(pkt.frame)
    assert got_frames == list(range(0, ntrig - 1))

    print(content)



import datetime
import time
import uuid
from enum import Enum
from typing import NewType, Literal, Annotated, Optional, TypeAlias, Iterable, Any

from pydantic import UUID4, BaseModel, validate_call, UrlConstraints, Field, TypeAdapter

from uuid import uuid4
from functools import cache, cached_property

from pydantic_core import Url

ZmqUrl = Annotated[Url, UrlConstraints(allowed_schemes=["tcp"])]

StreamName = NewType("StreamName", str)
"""
strongly typed stream name (str)
"""
WorkerName = NewType("WorkerName", str)
_WorkerTagT: TypeAlias = str
WorkerTag = NewType("WorkerTag", _WorkerTagT)
"""
Strongly typed worker tag (str)
"""
IngesterName = NewType("IngesterName", str)
VirtualConstraint = NewType("VirtualConstraint", int)
"""
Stronly typed constraint for workers (int)
"""
EventNumber = NewType("EventNumber", int)
"""
strongly typed event number (int)
"""

ParameterName = NewType("ParameterName", str)

HashDigest = NewType("HashDigest", str)

GENERIC_WORKER = WorkerTag("generic")

MappingName = NewType("MappingName", str)


class VirtualWorker(BaseModel):
    """
    virtual worker with a number and tags

    Attributes:
        tags: set of tags which a worker must have to get this event
        constraint: a VirtualConstraint to which worker this event should be delivered, if None, deliver to all workers with matching tags
    """

    tags: set[WorkerTag] = {GENERIC_WORKER}
    constraint: Optional[VirtualConstraint] = None


class RedisKeys:
    PREFIX = "dranspose"

    @staticmethod
    @cache
    @validate_call
    def config(
        typ: Literal["ingester", "worker", "reducer", "*"] = "*",
        instance: IngesterName | WorkerName | Literal["reducer", "*"] = "*",
    ) -> str:
        if typ == "reducer":
            instance = "reducer"
        return f"{RedisKeys.PREFIX}:{typ}:{instance}:config"

    @staticmethod
    @cache
    @validate_call
    def ready(uuid: Optional[UUID4 | Literal["*"]] = None) -> str:
        return f"{RedisKeys.PREFIX}:ready:{uuid}"

    @staticmethod
    @cache
    @validate_call
    def assigned(uuid: Optional[UUID4 | Literal["*"]] = None) -> str:
        return f"{RedisKeys.PREFIX}:assigned:{uuid}"

    @staticmethod
    @cache
    @validate_call
    def clock(uuid: Optional[UUID4 | Literal["*"]] = None) -> str:
        return f"{RedisKeys.PREFIX}:clock:{uuid}"

    @staticmethod
    @cache
    @validate_call
    def lock() -> str:
        return f"{RedisKeys.PREFIX}:controller_lock"

    @staticmethod
    @cache
    def updates() -> str:
        return f"{RedisKeys.PREFIX}:controller:updates"

    @staticmethod
    @cache
    @validate_call
    def parameters(
        name: str,
        uuid: UUID4 | Literal["*"],
    ) -> str:
        return f"{RedisKeys.PREFIX}:parameters:{name}:{uuid}"

    @staticmethod
    @cache
    @validate_call
    def parameter_description(
        name: str | Literal["*"] = "*",
    ) -> str:
        return f"{RedisKeys.PREFIX}:parameter_description:{name}"


class ProtocolException(Exception):
    pass


class ControllerUpdate(BaseModel):
    mapping_uuid: UUID4
    parameters_version: dict[ParameterName, UUID4]
    target_parameters_hash: Optional[HashDigest] = None
    active_streams: list[StreamName] = []
    finished: bool = False


class WorkParameter(BaseModel):
    name: str
    data: bytes
    value: Optional[int | str | bytes | float | bool] = None
    uuid: UUID4 = Field(default_factory=uuid4)

    def __repr_args__(self) -> Iterable[tuple[str | None, Any]]:
        args = super().__repr_args__()
        new_args = []
        for name, val in args:
            new_val = val
            if name == "data" and len(val) > 100:
                new_val = (
                    val[:50]
                    + f"... (total length {len(val)} bytes) ...".encode()
                    + val[-50:]
                )
            if name == "value" and isinstance(val, str) and len(val) > 100:
                new_val = (
                    val[:50]
                    + f"... (total length {len(val)} characters) ..."
                    + val[-50:]
                )
            if name == "value" and isinstance(val, bytes) and len(val) > 100:
                new_val = (
                    val[:50]
                    + f"... (total length {len(val)} bytes) ...".encode()
                    + val[-50:]
                )
            new_args.append((name, new_val))
        return new_args


class WorkAssignment(BaseModel):
    event_number: EventNumber
    assignments: dict[StreamName, list[WorkerName]]

    def get_workers_for_streams(self, streams: list[StreamName]) -> "WorkAssignment":
        ret = WorkAssignment(event_number=self.event_number, assignments={})
        for stream in streams:
            if stream in self.assignments:
                ret.assignments[stream] = self.assignments[stream]
        return ret

    def get_all_workers(self) -> set[WorkerName]:
        return set([x for stream in self.assignments.values() for x in stream])


WorkAssignmentList = TypeAdapter(list[WorkAssignment])


class DistributedStateEnum(Enum):
    READY = "ready"
    IDLE = "idle"
    FINISHED = "finished"


class WorkerTimes(BaseModel):
    get_assignments: float = 0.0
    get_messages: float = 0.0
    assemble_event: float = 0.0
    custom_code: float = 0.0
    send_result: float = 0.0
    no_events: int = 1

    @classmethod
    def from_timestamps(
        cls,
        start: float,
        assignments: float,
        messages: float,
        event: float,
        custom: float,
        send: float,
    ) -> "WorkerTimes":
        return WorkerTimes(
            get_assignments=assignments - start,
            get_messages=messages - assignments,
            assemble_event=event - messages,
            custom_code=custom - event,
            send_result=send - custom,
        )

    def __add__(self, other: "WorkerTimes") -> "WorkerTimes":
        return WorkerTimes(
            get_assignments=self.get_assignments + other.get_assignments,
            get_messages=self.get_messages + other.get_messages,
            assemble_event=self.assemble_event + other.assemble_event,
            custom_code=self.custom_code + other.custom_code,
            send_result=self.send_result + other.send_result,
            no_events=self.no_events + other.no_events,
        )

    @cached_property
    def total(self) -> float:
        return self.get_assignments + self.get_messages + self.active

    @cached_property
    def active(self) -> float:
        return self.assemble_event + self.custom_code + self.send_result

    @cached_property
    def load(self) -> float:
        return self.active / self.total


class IntervalLoad(BaseModel):
    total: float
    active: float
    events: int

    @cached_property
    def load(self) -> float:
        return self.active / self.total


class WorkerLoad(BaseModel):
    last_event: int
    intervals: dict[int | Literal["scan"], IntervalLoad]


SystemLoadType = dict[WorkerName, WorkerLoad]


class BaseUpdate(BaseModel):
    state: DistributedStateEnum


class WorkerUpdate(BaseUpdate):
    source: Literal["worker"] = "worker"
    completed: list[EventNumber] = []
    worker: WorkerName
    has_result: list[bool] = []
    processing_times: Optional[WorkerTimes] = None


class ReducerUpdate(BaseUpdate):
    source: Literal["reducer"] = "reducer"
    completed: Optional[EventNumber] = None
    worker: Optional[WorkerName] = None


class IngesterUpdate(BaseUpdate):
    source: Literal["ingester"] = "ingester"
    ingester: IngesterName


DistributedUpdateType = WorkerUpdate | ReducerUpdate | IngesterUpdate

DistributedUpdate = TypeAdapter(WorkerUpdate | ReducerUpdate | IngesterUpdate)


class BuildGitMeta(BaseModel):
    commit_hash: HashDigest
    branch_name: str
    timestamp: datetime.datetime
    repository_url: Url


class DistributedState(BaseModel):
    service_uuid: UUID4 = Field(default_factory=uuid.uuid4)
    mapping_uuid: Optional[UUID4] = None
    dranspose_version: Optional[str] = None
    mapreduce_version: Optional[BuildGitMeta] = None
    parameters_hash: Optional[HashDigest] = None
    processed_events: int = 0
    event_rate: float = 0.0

    # write a h5 dump function for metadata


class ConnectedWorker(BaseModel):
    name: WorkerName
    service_uuid: UUID4
    last_seen: float = Field(default_factory=time.time)


class IngesterState(DistributedState):
    name: IngesterName
    url: ZmqUrl
    connected_workers: dict[UUID4, ConnectedWorker] = {}
    streams: list[StreamName] = []


class WorkerState(DistributedState):
    name: WorkerName
    ingesters: list[IngesterState] = []
    tags: set[WorkerTag] = {GENERIC_WORKER}


class ReducerState(DistributedState):
    name: str = "reducer"
    url: ZmqUrl


class EnsembleState(BaseModel):
    ingesters: list[IngesterState]
    workers: list[WorkerState]
    reducer: Optional[ReducerState]
    controller_version: Optional[str] = None
    parameters_version: dict[ParameterName, UUID4] = {}
    parameters_hash: Optional[HashDigest] = None

    def get_streams(self) -> list[StreamName]:
        ingester_streams = set([s for i in self.ingesters for s in i.streams])
        worker_streams = [
            set([s for i in w.ingesters for s in i.streams]) for w in self.workers
        ]

        return list(ingester_streams.intersection(*worker_streams))

    def get_workers(self) -> list[WorkerName]:
        return [w.name for w in self.workers]



import json
import logging
from typing import Callable, Awaitable, Optional

import aiohttp
import pytest
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.protocol import (
    StreamName,
)
from tests.utils import wait_for_controller

payload = """{"streams": ["albaem-01_ch1", "albaem-01_ch3", "orca", "pcap_rot", "pcap_trigts", "albaem-xrd_ch1"], "scan": {"_name": "timescan", "_in_pars": [10, 0.03, 0.03], "_out_pars": null, "_aborted": false, "_stopped": false, "_processingStop": false, "_parent_macro": null, "_executor": "<<non-serializable: MacroExecutor>>", "_macro_line": "timescan(10, 0.03, 0.03) -> 39d8ecd6-a0e9-11ee-96fa-0050569cfe6d", "_interactive_mode": true, "_macro_thread": "<<non-serializable: OmniWorker>>", "_id": "39d8ecd6-a0e9-11ee-96fa-0050569cfe6d", "_desc": "Macro 'timescan(10, 0.03, 0.03) -> 39d8ecd6-a0e9-11ee-96fa-0050569cfe6d'", "_macro_status": {"id": "39d8ecd6-a0e9-11ee-96fa-0050569cfe6d", "range": [0.0, 100.0], "state": "start", "step": 0.0}, "_pause_event": "<<non-serializable: PauseEvent>>", "inited_class_list": ["<<non-serializable: type>>"], "log_name": "Macro[timescan]", "log_full_name": "B304A/DOOR/01.Macro[timescan]", "log_obj": "<<non-serializable: Logger>>", "log_handlers": [], "log_parent": "<<non-serializable: weakref>>", "log_children": {"140150443868416": "<<non-serializable: weakref>>", "140150443868848": "<<non-serializable: weakref>>", "140150443868464": "<<non-serializable: weakref>>", "140150443868896": "<<non-serializable: weakref>>", "140150443868608": "<<non-serializable: weakref>>"}, "_hooks": [["<<non-serializable: ExecMacroHook>>", ["pre-scan"]], ["<<non-serializable: ExecMacroHook>>", ["pre-scan"]], ["<<non-serializable: ExecMacroHook>>", ["pre-scan"]], ["<<non-serializable: ExecMacroHook>>", ["pre-scan"]], ["<<non-serializable: ExecMacroHook>>", ["post-scan"]], ["<<non-serializable: ExecMacroHook>>", ["post-scan"]], ["<<non-serializable: ExecMacroHook>>", ["post-scan"]], ["<<non-serializable: ExecMacroHook>>", ["post-scan"]], ["<<non-serializable: ExecMacroHook>>", ["pre-scan"]]], "_hookHintsDict": {"_ALL_": ["<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>"], "_NOHINTS_": [], "pre-scan": ["<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>"], "post-scan": ["<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>", "<<non-serializable: ExecMacroHook>>"]}, "nb_points": 10, "integ_time": 0.03, "latency_time": 0.03, "_gScan": "<<non-serializable: TScan>>", "_data": {}}}"""


@pytest.mark.asyncio
async def test_not_enough_workers(
    controller: None,
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    reducer: Callable[[Optional[str]], Awaitable[None]],
) -> None:
    await reducer(None)
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("orca")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )
    await wait_for_controller(streams={StreamName("orca")})
    async with aiohttp.ClientSession() as session:
        data = json.loads(payload)
        resp = await session.post(
            "http://localhost:5000/api/v1/sardana_hook",
            json=data,
        )
        assert resp.status == 200
        await resp.json()

        ma = await session.get("http://localhost:5000/api/v1/mapping")
        mapping = await ma.json()
        logging.info("map in %s", mapping)

        assert mapping == {
            "parts": {
                "sardana": {
                    "mapping": {
                        "orca": [
                            [{"tags": ["generic"], "constraint": 0}],
                            [{"tags": ["generic"], "constraint": 1}],
                            [{"tags": ["generic"], "constraint": 2}],
                            [{"tags": ["generic"], "constraint": 3}],
                            [{"tags": ["generic"], "constraint": 4}],
                            [{"tags": ["generic"], "constraint": 5}],
                            [{"tags": ["generic"], "constraint": 6}],
                            [{"tags": ["generic"], "constraint": 7}],
                            [{"tags": ["generic"], "constraint": 8}],
                            [{"tags": ["generic"], "constraint": 9}],
                        ]
                    }
                },
                "reserved_start_end": {
                    "mapping": {"orca": [[{"tags": ["generic"], "constraint": None}]]}
                },
            },
            "sequence": ["reserved_start_end", "sardana", "reserved_start_end"],
        }



import logging

from dranspose.data.eiger_legacy import EigerLegacyImage
from dranspose.event import EventData
from dranspose.middlewares.eiger_legacy import parse
from dranspose.protocol import StreamName

logger = logging.getLogger(__name__)


class LegacyWorker:
    def __init__(self, *args, **kwargs):
        pass

    def process_event(self, event: EventData, parameters=None, *args, **kwargs):
        if "eiger" in event.streams:
            data = parse(event.streams[StreamName("eiger")])
            if isinstance(data, EigerLegacyImage):
                data.buffer = b"some data removed"
            logger.info("parsed packet %s", data)



import itertools
import time

import numpy as np

from dranspose.event import InternalWorkerMessage

from dranspose.data.stream1 import Stream1Start, Stream1Data, Stream1End


class SlowSource:
    def __init__(self):
        pass

    def get_source_generators(self):
        return [
            self.pilatus_source(),
        ]

    def pilatus_source(self):
        msg_number = itertools.count(0)

        stins_start = Stream1Start(filename="", msg_number=next(msg_number))
        start = InternalWorkerMessage(
            event_number=0,
            streams={"pilatus": stins_start.to_stream_data()},
        )
        yield start

        frameno = 0
        # for image in self.fd["/entry/measurement/pilatus/frames"]:
        for _ in range(10):
            image = np.ones((10, 10))
            # dat = compress_lz4(image)
            stins = Stream1Data(
                msg_number=next(msg_number),
                frame=frameno,
                shape=image.shape,
                compression="none",  # compression="bslz4"
                type=str(image.dtype),
                data=image,
            )
            img = InternalWorkerMessage(
                event_number=frameno + 1,
                streams={"pilatus": stins.to_stream_data()},
            )
            yield img
            frameno += 1
            time.sleep(1)

        stins_end = Stream1End(msg_number=next(msg_number))
        end = InternalWorkerMessage(
            event_number=frameno + 1,
            streams={"pilatus": stins_end.to_stream_data()},
        )
        yield end



from __future__ import annotations
from datetime import datetime
from typing import Optional

import zmq
from pydantic import BaseModel, TypeAdapter

from dranspose.event import StreamData


class PositionCapField(BaseModel):
    name: str
    type: str
    value: Optional[int | float] = None

    def parse(self, value: str) -> PositionCapField:
        ret = PositionCapField(type=self.type, name=self.name)
        if self.type == "uint32":
            ret.value = int(value)
        elif self.type == "double":
            ret.value = float(value)
        else:
            raise Exception("unknown data")
        return ret


class PositionCapStart(BaseModel):
    arm_time: datetime

    def to_stream_data(self, fields: list[PositionCapField]) -> StreamData:
        data = f"""arm_time: {self.arm_time.strftime('%Y-%m-%dT%H:%M:%S.%fZ')}
missed: 0
process: Scaled
format: ASCII
fields:
"""
        for field in fields:
            where = field.name.split(".")[-1]
            data += f" {field.name[:-len(where) - 1]} {field.type} {where}\n"
        return StreamData(typ="PCAP_RAW", frames=[zmq.Frame(data.encode())])


class PositionCapValues(BaseModel):
    fields: dict[str, PositionCapField] = {}

    def to_stream_data(self) -> StreamData:
        data = " " + " ".join(map(lambda f: str(f.value), self.fields.values())) + "\n"
        return StreamData(typ="PCAP_RAW", frames=[zmq.Frame(data.encode())])


class PositionCapEnd(BaseModel):
    def to_stream_data(self) -> StreamData:
        data = "END 0 Disarmed\n"
        return StreamData(typ="PCAP_RAW", frames=[zmq.Frame(data.encode())])


PositionCapPacketType = PositionCapStart | PositionCapValues | PositionCapEnd

PositionCapPacket = TypeAdapter(PositionCapStart | PositionCapValues | PositionCapEnd)
"""
A union type for PCAP_RAW packets
"""



from enum import IntEnum
from typing import Any, List, Literal

from numpy.typing import NDArray
import zmq
from pydantic import BaseModel, TypeAdapter, ConfigDict

from dranspose.event import StreamData

LECROY_TYPE = "lecroy_9fwdWp6Rq7"


class WhatEnum(IntEnum):
    PREPARE = 0
    START = 1
    SEQEND = 2
    STOP = 3


class LecroyBase(BaseModel):
    frame: int

    def to_stream_data(self) -> StreamData:
        dat = self.model_dump_json()
        return StreamData(typ=LECROY_TYPE, frames=[zmq.Frame(dat)])


class LecroyPrepare(LecroyBase):
    # len(parts) parts[0]
    # 1 b'{"htype": "msg", "what": 0, "frame": 0}'
    """
    Example:
        ``` py
        LecroyStart(
            htype='msg'
            what=0,
            frame=0,
        )
        ```
    """

    htype: Literal["msg"]
    what: Literal[WhatEnum.PREPARE]  # PREPARE = 0


class LecroySeqStart(LecroyBase):
    # len(parts) parts[0]
    # 1 b'{"htype": "msg", "what": 1, "frame": 0, "ntriggers": -1, "seqno": 0, "channels": [2, 4]}'
    """
    Example:
        ``` py
        LecroyStart(
            htype='msg'
            what=1,
            frame=0,
            ntriggers=-1
            seqno=0
            channels=[2, 4]
        )
        ```
    """

    htype: Literal["msg"]
    what: Literal[WhatEnum.START]  # START = 1
    ntriggers: int
    seqno: int
    channels: List[int]


class LecroySeqEnd(LecroyBase):
    # len(parts) parts[0]
    # 1 b'{"htype": "msg", "what": 2, "frame": 2}'
    """
    Example:
        ``` py
        LecroySeqEnd(
            htype='msg'
            what=2,
            frame=2,
        )
        ```
    """

    htype: Literal["msg"]
    what: Literal[WhatEnum.SEQEND]  # SEQEND = 2


class LecroyEnd(LecroyBase):
    # len(parts) parts[0]
    """
    Example:
        ``` py
        LecroyEnd(
            htype='msg'
            what=3,
            frame=66,
            frames=66,
        )
        ```
    """

    frames: int
    what: Literal[WhatEnum.STOP]  # STOP = 3


class LecroyData(LecroyBase):
    # len(parts) parts[0]
    # 3 b'{"htype": "traces", "ch": 2, "ts": 1740563614.969933, "frame": 0, "shape": [1, 8002], "horiz_offset": -1.0000505879544622e-07, "horiz_interval": 1.25000001668929e-11, "dtype": "float64"}'
    """
    Each "traces" message has 3 zmq parts:
    metadata (json), waveforms (np.array), timestamps (list)

    Example:
        ``` py
        LecroyData(
            htype='traces',
            ch: 2,
            ts: 1740563614.969933,
            frame: 0,
            shape: [1, 8002],
            horiz_offset: -1.0000505879544622e-07,
            horiz_interval: 1.25000001668929e-11,
            dtype: "float64"
            )
        ```
    """

    htype: Literal["traces"]
    ch: int
    ts: float
    frame: int
    shape: List[int]
    dtype: str
    horiz_offset: float
    horiz_interval: float


class LecroyParsed(LecroySeqStart):
    # tells Pydantic to allow arbitrary Python types (like NumPy arrays)
    # without trying to validate them strictly
    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    meta: List[LecroyData] = []
    data: List[NDArray[Any]] = []
    timestamps: List[List[float]] = []


LecroyPacket: TypeAdapter = TypeAdapter(LecroyPrepare | LecroySeqStart | LecroyEnd | LecroySeqEnd | LecroyData | LecroyParsed)  # type: ignore [type-arg]
"""
Union type for Lecroy packets
"""



from typing import Any

from dranspose.event import EventData


class TestWorker:
    def __init__(self, **kwargs: Any) -> None:
        pass

    def process_event(self, event: EventData, *args, **kwargs):
        for stream in event.streams:
            for i, _ in enumerate(event.streams[stream].frames):
                event.streams[stream].frames[i] = event.streams[stream].frames[i][:5000]
        return event, args, kwargs



from typing import AsyncGenerator, Optional

import zmq

from dranspose.data.pcap import PCAPPacket, PCAPStart, PCAPData, PCAPEnd
from dranspose.event import StreamData
from dranspose.ingester import Ingester, IngesterSettings
from dranspose.protocol import StreamName, ZmqUrl


class ZmqSubPCAPSettings(IngesterSettings):
    upstream_url: ZmqUrl


class ZmqSubPCAPIngester(Ingester):
    def __init__(self, settings: Optional[ZmqSubPCAPSettings] = None) -> None:
        if settings is not None:
            self._streaming_PCAP_settings = settings
        else:
            self._streaming_PCAP_settings = ZmqSubPCAPSettings()

        super().__init__(settings=self._streaming_PCAP_settings)
        self.in_socket: Optional[zmq._future._AsyncSocket] = None

    async def run_source(self, stream: StreamName) -> AsyncGenerator[StreamData, None]:
        self.in_socket = self.ctx.socket(zmq.SUB)
        self.in_socket.connect(str(self._streaming_PCAP_settings.upstream_url))
        self.in_socket.setsockopt(zmq.SUBSCRIBE, b"")
        self._logger.info(
            "subscribed to %s", self._streaming_PCAP_settings.upstream_url
        )

        while True:
            self._logger.debug("clear up insocket")
            parts = await self.in_socket.recv_multipart(copy=False)
            try:
                packet = PCAPPacket.validate_json(parts[0].bytes)
            except Exception as e:
                self._logger.warning("packet not valid %s", e.__repr__())
                continue
            self._logger.debug("received frame with header %s", packet)
            if type(packet) is PCAPStart:
                self._logger.info("start of new sequence %s", packet)
                yield StreamData(typ="PCAP", frames=parts)
                break
        while True:
            parts = await self.in_socket.recv_multipart(copy=False)
            try:
                packet = PCAPPacket.validate_json(parts[0].bytes)
            except Exception as e:
                self._logger.error("packet not valid %s", e.__repr__())
                continue
            if isinstance(packet, PCAPData):
                if len(parts) != 1:
                    self._logger.error("PCAP multipart messages are not supported")
                    continue
                self._logger.debug("received data is %s", packet)
                vect_keys = {k: len(v) for k, v in iter(packet) if isinstance(v, list)}
                self._logger.debug("vect_keys are %s", vect_keys)
                lenghts = set(vect_keys.values())
                if len(lenghts) != 1:
                    self._logger.error("All lists in msg must have the same length")
                    continue
                for i in range(lenghts.pop()):
                    frame = {}
                    for k, v in iter(packet):
                        if k in vect_keys.keys():
                            frame[k] = v[i]
                        elif k == "frame_number":
                            frame[k] = v + i  # Increment frame_number
                        else:
                            frame[k] = v  # Keep the original scalar value
                    new_packet = PCAPData(**frame)
                    self._logger.debug("send out newly created package %s", new_packet)
                    frame_bytes = new_packet.model_dump_json().encode()
                    yield StreamData(typ="PCAP", frames=[frame_bytes])
            elif isinstance(packet, PCAPEnd):
                self._logger.info("reached end %s", packet)
                yield StreamData(typ="PCAP", frames=parts)
                break
        while True:
            self._logger.debug("discarding messages until next run")
            await self.in_socket.recv_multipart(copy=False)

    async def stop_source(self, stream: StreamName) -> None:
        if self.in_socket:
            self._logger.info("closing socket without linger")
            self.in_socket.close(linger=0)
            self.in_socket = None



import logging
from datetime import timedelta

from dranspose.data.positioncap import PositionCapValues
from dranspose.event import EventData
from dranspose.middlewares.positioncap import PositioncapParser

logger = logging.getLogger(__name__)


class PcapWorker:
    def __init__(self, *args, **kwargs):
        self.pcap = PositioncapParser()
        pass

    def process_event(self, event: EventData, *args, **kwargs):
        if "pcap" in event.streams:
            res = self.pcap.parse(event.streams["pcap"])
            if isinstance(res, PositionCapValues):
                triggertime = timedelta(seconds=res.fields["PCAP.TS_TRIG.Value"].value)
                logger.info(
                    "got values %s at timestamp %s",
                    res,
                    self.pcap.arm_time + triggertime,
                )



import asyncio
import datetime
import logging
from typing import Awaitable, Callable, Optional, Any

import aiohttp
import numpy as np

import pytest
import zmq.asyncio
import zmq
from psutil._common import snicaddr, snetio

from dranspose.workload_generator import Statistics, NetworkConfig


async def consume(
    ctx: zmq.Context[Any], num: int, typ: int = zmq.PULL, port: int = 9999
) -> int:
    s = ctx.socket(typ)
    logging.info("created socket")
    s.connect(f"tcp://127.0.0.1:{port}")
    if typ == zmq.SUB:
        s.setsockopt(zmq.SUBSCRIBE, b"")
    logging.info("connected socket to port %s", port)
    for _ in range(num):
        data = await s.recv_multipart(copy=False)
        logging.debug("received data %s", data)

    return num


@pytest.mark.parametrize("srv,cli", [(zmq.PUSH, zmq.PULL), (zmq.PUB, zmq.SUB)])
@pytest.mark.asyncio
async def test_debugger(
    workload_generator: Callable[[Optional[int]], Awaitable[None]], srv: int, cli: int
) -> None:
    await workload_generator(5003)

    nframes = 5

    ctx = zmq.asyncio.Context()
    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5003/api/v1/config")
        state = NetworkConfig.model_validate_json(await st.read())
        for ifname, ifaddrs in state.addresses.items():
            logging.info("interface %s : %s", ifname, state.stats[ifname])
            for ifaddr in ifaddrs:
                logging.info("    %s", snicaddr(*ifaddr))

        st = await session.get("http://localhost:5003/api/v1/finished")
        state = await st.json()
        logging.info("gen state %s", state)

        task = asyncio.create_task(consume(ctx, nframes + 2, typ=cli))
        logging.info("created consumer task")

        st = await session.post(
            "http://localhost:5003/api/v1/open_socket",
            json={"type": srv, "port": 9999},
        )
        state = await st.json()
        logging.info("open %s", state)

        await asyncio.sleep(1)

        st = await session.post(
            "http://localhost:5003/api/v1/frames",
            json={"number": nframes, "time": 0.0001, "shape": [100, 100]},
        )
        state = await st.json()
        st.close()
        logging.info("sending frames %s", state)

        st = await session.get("http://localhost:5003/api/v1/finished")
        state = await st.json()
        st.close()
        while not state:
            await asyncio.sleep(0.5)
            st = await session.get("http://localhost:5003/api/v1/finished")
            state = await st.json()
            st.close()

            st = await session.get("http://localhost:5003/api/v1/statistics")
            stat = await st.json()
            st.close()
            logging.info("fps %s", stat["fps"])

        st = await session.post("http://localhost:5003/api/v1/close_socket")
        state = await st.json()
        st.close()
        logging.info("close %s", state)

        st = await session.get("http://localhost:5003/api/v1/statistics")
        stat = Statistics.model_validate_json(await st.read())
        st.close()
        last = {}
        last_sent = 0
        for t, sent, stats in stat.snapshots:
            logging.info(
                "timestamp %s: sent %d, Δ%d",
                datetime.datetime.fromtimestamp(t),
                sent,
                sent - last_sent,
            )
            last_sent = sent
            for ifname, nst in stats.items():
                if ifname not in last:
                    last[ifname] = nst
                    logging.info("    %s: %s", ifname, nst)
                else:
                    delta = snetio(*map(lambda x: x[1] - x[0], zip(last[ifname], nst)))
                    logging.info("   Δ%s: %s", ifname, delta)
                    last[ifname] = nst
    await task
    ctx.destroy()


@pytest.mark.asyncio
async def test_sink(
    workload_generator: Callable[[Optional[int]], Awaitable[None]]
) -> None:
    await workload_generator(5003)
    await workload_generator(5004)

    async with aiohttp.ClientSession() as session:
        nframes = 100

        st = await session.post(
            "http://localhost:5003/api/v1/open_socket",
            json={"type": zmq.PUSH, "port": 9999},
        )
        state = await st.json()
        logging.info("open %s", state)

        st = await session.post(
            "http://localhost:5004/api/v1/connect_socket",
            json={"type": zmq.PULL, "url": "tcp://127.0.0.1:9999"},
        )
        state = await st.json()
        logging.info("open %s", state)

        await asyncio.sleep(0.4)

        st = await session.post(
            "http://localhost:5003/api/v1/frames",
            json={"number": nframes, "time": 0.01, "shape": [100, 100]},
        )
        state = await st.json()
        logging.info("sending frames %s", state)

        st = await session.get("http://localhost:5003/api/v1/finished")
        state = await st.json()
        logging.info("is finished %s", state)
        while not state:
            await asyncio.sleep(0.5)
            st = await session.get("http://localhost:5003/api/v1/finished")
            state = await st.json()

            st = await session.get("http://localhost:5003/api/v1/statistics")
            stat = await st.json()
            logging.info("sender fps %s", stat["fps"])

            st = await session.get("http://localhost:5004/api/v1/statistics")
            stat = await st.json()
            logging.info("receiver fps %s", stat["fps"])
            logging.info(
                "receiver latencies mean %f, max %f",
                np.mean(stat["deltas"]),
                max(stat["deltas"]),
            )

        st = await session.post("http://localhost:5003/api/v1/close_socket")
        state = await st.json()
        logging.info("close source %s", state)

        st = await session.post("http://localhost:5004/api/v1/close_socket")
        state = await st.json()
        logging.info("close sink %s", state)



import logging
from asyncio import StreamReader, StreamWriter
from typing import Callable, Coroutine

import pytest
import asyncio

from dranspose.helpers.utils import cancel_and_wait


async def handle_client(reader: StreamReader, writer: StreamWriter) -> None:
    request = (await reader.read(255)).decode("utf8")
    if request.strip() != "":
        logging.error("bad initial hello %s", request)
    writer.write(b"OK\n")
    await writer.drain()
    for i in range(1000):
        writer.write(b"testdata" * 1000 + b"\n")
        await writer.drain()
        if i % 100 == 0:
            logging.info("sent out %d", i)
        await asyncio.sleep(0.001)

    await writer.drain()
    writer.close()


@pytest.mark.skip
@pytest.mark.asyncio
async def test_tcp(stream_pcap: Callable[[int], Coroutine[None, None, None]]) -> None:
    server = await asyncio.start_server(handle_client, "localhost", 8889)
    task = asyncio.create_task(server.serve_forever())

    reader, writer = await asyncio.open_connection("127.0.0.1", 8889)
    writer.write(b"\n")
    data = await reader.readline()
    assert data == b"OK\n"
    logging.info("connected to %s", "127.0.0.1")

    for i in range(1000):
        line = await reader.readline()
        if line == b"":
            break
        logging.debug("line is %s", line)

    await cancel_and_wait(task)



import asyncio
import logging

import pytest


import redis.asyncio as redis


@pytest.mark.asyncio
async def test_distributed() -> None:
    r = redis.Redis(host="localhost", port=6379, decode_responses=True, protocol=3)

    keys = await r.keys("dranspose:*")
    logging.info("keys %s", keys)

    key = await r.set("testkey", "didit", px=800, nx=True)
    logging.info("key set to %s", key)
    assert key is True
    key = await r.set("testkey", "didit", px=500, nx=True)
    logging.info("key set to %s", key)
    assert key is None

    ttl = await r.pttl("testkey")
    logging.info("ttl is %f", ttl)
    await asyncio.sleep(1)

    key = await r.set("testkey", "didit", px=500, nx=True)
    logging.info("key set to %s", key)
    assert key is True

    await r.aclose()



from typing import AsyncGenerator, Optional

import zmq

from dranspose.data.lecroy import (
    LECROY_TYPE,
    LecroyPacket,
    LecroyPrepare,
    LecroySeqEnd,
    LecroySeqStart,
    LecroyData,
    LecroyEnd,
)
from dranspose.event import StreamData
from dranspose.ingester import Ingester, IngesterSettings
from dranspose.protocol import StreamName, ZmqUrl


class ZmqSubLecroySettings(IngesterSettings):
    upstream_url: ZmqUrl


class ZmqSubLecroyIngester(Ingester):
    """
    In this protocol a frame does not correspond to a trigger
    if there are nch channels, a trigger will produce nch "traces" messages.
    There are 2 moddes

    Continuous mode
    m0, m1, tt..., m2, tt..., m2, tt..., m2, [...], m3
    that is
        LecroyPrepare - LecroySeqStart
    followed by a lot of
        LecroyData - LecroyData... - LecroySeqEnd
    and ended by a
        LecroyEnd

    Sequential mode
    m0, m1, tt..., m2, m1, tt..., m2, m1, tt..., m2, [...],  m3
    that is
        LecroyPrepare
    followed by a lot of
        LecroySeqStart - LecroyData - LecroyData... - LecroySeqEnd
    and ended by a
        LecroyEnd

    Notes: I decided to cache the first LecroySeqStart received and to send it
    over and over at the beginning of each StreamData in continuos_mode
    so that the message structure would always be:
    LecroySeqStart - LecroyData... - LecroySeqEnd
    for both modes.
    """

    def __init__(self, settings: Optional[ZmqSubLecroySettings] = None) -> None:
        if settings is not None:
            self._streaming_Lecroy_settings = settings
        else:
            self._streaming_Lecroy_settings = ZmqSubLecroySettings()

        super().__init__(settings=self._streaming_Lecroy_settings)
        self.in_socket: Optional[zmq._future._AsyncSocket] = None

    async def run_source(self, stream: StreamName) -> AsyncGenerator[StreamData, None]:
        self.in_socket = self.ctx.socket(zmq.SUB)
        self.in_socket.connect(str(self._streaming_Lecroy_settings.upstream_url))
        self.in_socket.setsockopt(zmq.SUBSCRIBE, b"")
        self._logger.info(
            "subscribed to %s", self._streaming_Lecroy_settings.upstream_url
        )

        while True:
            self._logger.debug("clear up insocket")
            parts = await self.in_socket.recv_multipart(copy=False)
            try:
                prepare_pkt = LecroyPacket.validate_json(parts[0].bytes)
            except Exception as e:
                self._logger.warning("packet not valid %s", e.__repr__())
                continue
            self._logger.debug(
                f"received frame of type {type(prepare_pkt)} with header {prepare_pkt}"
            )
            if isinstance(prepare_pkt, LecroyPrepare):
                self._logger.info("prepare new run %s", prepare_pkt)
                frames = parts
                parts = await self.in_socket.recv_multipart(copy=False)
                try:
                    seqstart_pkt = LecroyPacket.validate_json(parts[0].bytes)
                except Exception as e:
                    self._logger.warning("packet not valid %s", e.__repr__())
                    continue
                if isinstance(seqstart_pkt, LecroySeqStart):
                    self._logger.info("start of new sequence %s", seqstart_pkt)
                    yield StreamData(typ=LECROY_TYPE, frames=frames)
                    break
        continuos_mode = seqstart_pkt.ntriggers == -1
        frames = parts  # LecroySeqStart0
        while True:
            parts = await self.in_socket.recv_multipart(copy=False)
            try:
                packet = LecroyPacket.validate_json(parts[0].bytes)
            except Exception as e:
                self._logger.error("packet not valid %s", e.__repr__())
                continue
            self._logger.debug(
                f"received frame of type {type(packet)} with header {packet}"
            )
            # LecroySeqStart0 has been already received
            # receive LecroyData - LecroyData... - LecroySeqEnd
            # then send LecroySeqStart0 - LecroyData... - LecroySeqEnd
            if isinstance(packet, LecroySeqStart) or isinstance(packet, LecroyData):
                frames += parts
            elif isinstance(packet, LecroySeqEnd):
                frames += parts
                yield StreamData(typ=LECROY_TYPE, frames=frames)
                if continuos_mode:
                    frames = frames[:1]  # keep LecroySeqStart0
                else:
                    frames = []  # empty buffer
            elif isinstance(packet, LecroyEnd):
                self._logger.info("reached end %s", packet)
                if (continuos_mode and len(frames) != 1) or (
                    not continuos_mode and len(frames) != 0
                ):
                    self._logger.error(
                        "Untrasmitted frames left in the buffer %s",
                        frames.__repr__(),
                    )
                yield StreamData(typ=LECROY_TYPE, frames=parts)
                break

        while True:
            self._logger.debug("discarding messages until next run")
            await self.in_socket.recv_multipart(copy=False)

    async def stop_source(self, stream: StreamName) -> None:
        if self.in_socket:
            self._logger.info("closing socket without linger")
            self.in_socket.close(linger=0)
            self.in_socket = None



import asyncio
import os
import pickle
from typing import Awaitable, Callable, Any, Coroutine, Optional

import aiohttp

import pytest
import zmq.asyncio
import zmq
from pydantic_core import Url

from dranspose.data.stream1 import Stream1Data
from dranspose.event import EventData
from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.middlewares import stream1
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
    WorkerTag,
)

from dranspose.worker import Worker
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_debugger(
    controller: None,
    debug_worker: Callable[[Optional[str], Optional[list[str]]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
) -> None:
    await debug_worker("debugworker", ["debug"])
    assert "WORKER_TAGS" not in os.environ
    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5002/api/v1/last_events?number=1")
        content = await st.content.read()
        res: list[EventData] = pickle.loads(content)
        assert res == []

        st = await session.get("http://localhost:5002/api/v1/status")
        assert await st.json()

    await create_worker(WorkerName("w1"))

    state = await wait_for_controller(workers={WorkerName("w1")})

    assert set([w.name for w in state.workers]) == {"w1", "debugworker"}


@pytest.mark.asyncio
async def test_debug(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    debug_worker: Callable[[Optional[str], Optional[list[str]]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await debug_worker("debugworker", ["debug"])
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )

    state = await wait_for_controller(streams={StreamName("eiger")})
    assert set([w.name for w in state.workers]) == {"w1", "debugworker"}
    async with aiohttp.ClientSession() as session:
        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    if i % 4 != 0
                    else [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        ),
                        VirtualWorker(tags={WorkerTag("debug")}).model_dump(
                            mode="json"
                        ),
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        await resp.json()

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1))

    await wait_for_finish()

    context.destroy()

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5002/api/v1/last_events?number=2")
        content = await st.content.read()
        res: list[EventData] = pickle.loads(content)
        for ev in res:
            print("got event", ev.event_number, ev.streams.keys())
        pkg = stream1.parse(res[0].streams[StreamName("eiger")])
        print(pkg)
        if isinstance(pkg, Stream1Data):
            assert pkg.frame == 7
            assert pkg.shape == [1475, 831]
            assert pkg.type == "uint16"
        else:
            assert False



import asyncio
from typing import Awaitable, Callable, Any, Coroutine, Optional

import aiohttp

import pytest
import zmq.asyncio
import zmq
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)

from dranspose.worker import Worker
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_simple(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig - 5)
                ]
                + [None]
                + [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(ntrig - 5, ntrig)
                ],
            },
        )
        assert resp.status == 200

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1))

    content = await wait_for_finish()

    context.destroy()

    print(content)



import os
import pathlib
from pathlib import PosixPath

import asyncio
from typing import Awaitable, Callable, Any, Coroutine, Optional
import zmq.asyncio

import aiohttp
import pytest
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.tcp_positioncap import TcpPcapSettings
from dranspose.ingesters.zmqsub_contrast import (
    ZmqSubContrastIngester,
    ZmqSubContrastSettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)
from dranspose.worker import Worker, WorkerSettings
from examples.dummy.sw_trig_ingester import SoftTriggerPcapIngester
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_ingester(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_pcap: Callable[[int], Coroutine[None, None, None]],
    stream_pkls: Callable[
        [zmq.Context[Any], int, os.PathLike[Any] | str, float, int],
        Coroutine[Any, Any, None],
    ],
    tmp_path: pathlib.PurePath,
) -> None:
    await reducer("examples.dummy.reducer:FluorescenceReducer")
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w1"),
                worker_class="examples.dummy.worker:FluorescenceWorker",
            ),
        )
    )

    await create_ingester(
        ZmqSubContrastIngester(
            settings=ZmqSubContrastSettings(
                ingester_streams=[StreamName("contrast")],
                upstream_url=Url("tcp://localhost:5556"),
                ingester_url=Url("tcp://localhost:10000"),
            ),
        )
    )

    ntrig = 20

    await create_ingester(
        SoftTriggerPcapIngester(
            settings=TcpPcapSettings(
                ingester_streams=[StreamName("pcap"), StreamName("dummy")],
                upstream_url=Url("tcp://localhost:8889"),
                ingester_url=Url("tcp://localhost:10001"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("contrast"), StreamName("pcap")})
    async with aiohttp.ClientSession() as session:
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "contrast": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(ntrig)
                ],
                "pcap": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(ntrig)
                ],
                "dummy": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(ntrig)
                ],
            },
        )
        assert resp.status == 200
        await resp.json()

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_pcap(ntrig))

    asyncio.create_task(
        stream_pkls(
            context, 5556, PosixPath("tests/data/contrast-dump.pkls"), 0.001, zmq.PUB
        )
    )

    content = await wait_for_finish()

    context.destroy()

    print(content)



import logging
from typing import AsyncGenerator, Optional

import zmq

from dranspose.data.xspress3 import (
    XspressPacket,
    XspressStart,
    XspressImage,
    XspressEnd,
)
from dranspose.event import StreamData
from dranspose.ingester import Ingester, IngesterSettings
from dranspose.protocol import StreamName, ZmqUrl


class ZmqSubXspressSettings(IngesterSettings):
    upstream_url: ZmqUrl


class ZmqSubXspressIngester(Ingester):
    def __init__(self, settings: Optional[ZmqSubXspressSettings] = None) -> None:
        if settings is not None:
            self._streaming_xspress_settings = settings
        else:
            self._streaming_xspress_settings = ZmqSubXspressSettings()

        super().__init__(settings=self._streaming_xspress_settings)

        self.in_socket: Optional[zmq._future._AsyncSocket] = None

    async def run_source(self, stream: StreamName) -> AsyncGenerator[StreamData, None]:
        self.in_socket = self.ctx.socket(zmq.SUB)
        self.in_socket.connect(str(self._streaming_xspress_settings.upstream_url))
        self.in_socket.setsockopt(zmq.SUBSCRIBE, b"")
        self._logger.info(
            "subscribed to %s", self._streaming_xspress_settings.upstream_url
        )

        while True:
            self._logger.debug("clear up insocket")
            parts = await self.in_socket.recv_multipart(copy=False)
            if len(parts) != 1:
                self._logger.error(
                    "xpress3 never sends multipart messages, received %d parts",
                    len(parts),
                )
                continue
            try:
                packet = XspressPacket.validate_json(parts[0].bytes)
            except Exception as e:
                self._logger.warning("packet not valid %s", e.__repr__())
                continue
            self._logger.debug("received frame with header %s", packet)
            if type(packet) is XspressStart:
                self._logger.info("start of new sequence %s", packet)
                yield StreamData(typ="xspress", frames=parts)
                break
        while True:
            parts = await self.in_socket.recv_multipart(copy=False)
            try:
                packet = XspressPacket.validate_json(parts[0].bytes)
            except Exception as e:
                self._logger.error("packet parsable valid %s", e.__repr__())
                continue
            if type(packet) is XspressImage:
                if len(parts) == 3:
                    logging.debug("this is the new multipart format %d", len(parts))
                    yield StreamData(typ="xspress", frames=parts)
                else:
                    image = await self.in_socket.recv_multipart(copy=False)
                    meta = await self.in_socket.recv_multipart(copy=False)
                    yield StreamData(typ="xspress", frames=parts + image + meta)
            elif type(packet) is XspressEnd:
                yield StreamData(typ="xspress", frames=parts)
                break
        while True:
            self._logger.debug("discarding messages until next run")
            await self.in_socket.recv_multipart(copy=False)

    async def stop_source(self, stream: StreamName) -> None:
        if self.in_socket:
            self._logger.info("closing socket without linger")
            self.in_socket.close(linger=0)
            self.in_socket = None



from dranspose.event import ResultData
from dranspose.parameters import StrParameter, BinaryParameter


class FluorescenceReducer:
    def __init__(self, state=None, **kwargs):
        self.number = 0
        self.publish = {"map": {}}
        if state is not None and state.mapreduce_version is not None:
            self.publish["version"] = state.mapreduce_version.model_dump(mode="json")

    @staticmethod
    def describe_parameters():
        params = [
            StrParameter(name="string_parameter"),
            BinaryParameter(name="other_file_parameter"),
        ]
        return params

    def process_result(self, result: ResultData, parameters=None):
        print(result)
        if result.payload:
            self.publish["map"][result.payload["position"]] = result.payload[
                "concentations"
            ]

    def finish(self, parameters=None):
        print("finished dummy reducer work")



import asyncio
import logging
import os
import pickle
from pathlib import PosixPath
from typing import Callable, Optional, Awaitable, Any, Coroutine

import aiohttp
import numpy as np
import pytest
import zmq
from pydantic_core import Url


from dranspose.ingester import Ingester
from dranspose.ingesters.zmqsub_contrast import (
    ZmqSubContrastSettings,
    ZmqSubContrastIngester,
)
from dranspose.parameters import ParameterList
from dranspose.protocol import (
    WorkerName,
    StreamName,
    VirtualWorker,
    VirtualConstraint,
    EnsembleState,
)
from dranspose.worker import Worker, WorkerSettings
from tests.utils import wait_for_controller


@pytest.mark.asyncio
async def test_params(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_pkls: Callable[
        [zmq.Context[Any], int, os.PathLike[Any] | str, float, int],
        Coroutine[Any, Any, None],
    ],
) -> None:
    await reducer("examples.dummy.reducer:FluorescenceReducer")
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w1"),
                worker_class="examples.dummy.worker:FluorescenceWorker",
            ),
        )
    )

    await create_ingester(
        ZmqSubContrastIngester(
            settings=ZmqSubContrastSettings(
                ingester_streams=[StreamName("contrast")],
                upstream_url=Url("tcp://localhost:5556"),
                ingester_url=Url("tcp://localhost:10000"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("contrast")})
    async with aiohttp.ClientSession() as session:
        par = await session.get("http://localhost:5000/api/v1/parameters")
        assert par.status == 200
        params = ParameterList.validate_python(await par.json())

        logging.warning("params %s", params)

        resp = await session.post(
            "http://localhost:5000/api/v1/parameter/roi1",
            json=[0, 10],
        )
        assert resp.status == 200

        resp = await session.post(
            "http://localhost:5000/api/v1/parameter/additional",
            data=b"asdasd",
        )
        assert resp.status == 200
        hash = await resp.json()

        resp = await session.get(
            "http://localhost:5000/api/v1/parameter/additional",
        )
        assert resp.status == 200
        assert await resp.content.read() == b"asdasd"

        large = b"asd" * 1000
        resp = await session.post(
            "http://localhost:5000/api/v1/parameter/large",
            data=large,
        )
        assert resp.status == 200

        resp = await session.get(
            "http://localhost:5000/api/v1/parameter/large",
        )
        assert resp.status == 200
        assert await resp.content.read() == large

        resp = await session.get(
            "http://localhost:5000/api/v1/parameter/nonexistant",
        )
        assert resp.status == 404

        await asyncio.sleep(3)

        resp = await session.post(
            "http://localhost:5000/api/v1/parameter/additional",
            data=b"asdasd",
        )
        assert resp.status == 200

        resp = await session.post(
            "http://localhost:5000/api/v1/parameter/file_parameter",
            data=b"mydirectnumpyarray",
        )
        assert resp.status == 200

        arr = np.ones((10, 10))
        resp = await session.post(
            "http://localhost:5000/api/v1/parameter/file_parameter_file",
            data=pickle.dumps(arr),
        )
        assert resp.status == 200
        hash = await resp.json()

        st = await session.get("http://localhost:5000/api/v1/config")
        state = EnsembleState.model_validate(await st.json())
        print("state", state)
        while str(state.workers[0].parameters_hash) != hash:
            await asyncio.sleep(0.3)
            st = await session.get("http://localhost:5000/api/v1/config")
            state = EnsembleState.model_validate(await st.json())
            logging.warning("got state %s", state)

        ntrig = 20
        mp = {
            "contrast": [
                [VirtualWorker(constraint=VirtualConstraint(i)).model_dump(mode="json")]
                for i in range(ntrig)
            ],
        }
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json=mp,
        )
        assert resp.status == 200

        context = zmq.asyncio.Context()

        await stream_pkls(
            context, 5556, PosixPath("tests/data/contrast-dump.pkls"), 0.001, zmq.PUB
        )

        st = await session.get("http://localhost:5000/api/v1/progress")
        content = await st.json()
        while not content["finished"]:
            await asyncio.sleep(0.3)
            st = await session.get("http://localhost:5000/api/v1/progress")
            content = await st.json()

        context.destroy()



import asyncio
import logging
from contextlib import asynccontextmanager
from typing import AsyncGenerator, Optional, Any

import zmq
from fastapi import FastAPI, Body

from dranspose.data.sardana import (
    SardanaDataDescription,
    SardanaRecordEnd,
    SardanaRecordData,
    SardanaPacketType,
)
from dranspose.event import StreamData
from dranspose.helpers.utils import done_callback, cancel_and_wait
from dranspose.ingester import Ingester, IngesterSettings
from dranspose.protocol import StreamName


logger = logging.getLogger(__name__)


class HttpSardanaSettings(IngesterSettings):
    pass


class HttpSardanaIngester(Ingester):
    def __init__(self, settings: Optional[HttpSardanaSettings] = None) -> None:
        if settings is not None:
            self._http_sardana_settings = settings
        else:
            self._http_sardana_settings = HttpSardanaSettings()

        super().__init__(settings=self._http_sardana_settings)
        self.in_queue: Optional[asyncio.Queue[SardanaPacketType]] = asyncio.Queue()

    async def run_source(self, stream: StreamName) -> AsyncGenerator[StreamData, None]:
        self.in_queue = asyncio.Queue()
        while True:
            self._logger.debug("clear up insocket")
            packet = await self.in_queue.get()
            if isinstance(packet, SardanaDataDescription):
                self._logger.info("start of new sequence %s", packet)
                frame = zmq.Frame(packet.model_dump_json().encode("utf8"))
                yield StreamData(typ="sardana", frames=[frame])
                break
        while True:
            if self.in_queue is not None:
                packet = await self.in_queue.get()
            else:
                break
            frame = zmq.Frame(packet.model_dump_json().encode("utf8"))
            self._logger.debug("send stream data %s", frame.bytes)
            yield StreamData(typ="sardana", frames=[frame])
            if isinstance(packet, SardanaRecordEnd):
                break
        while True:
            self._logger.debug("discarding messages until next run")
            await self.in_queue.get()

    async def stop_source(self, stream: StreamName) -> None:
        if self.in_queue:
            self._logger.info("closing queue")
            self.in_queue = None

    async def receive(self, paket: SardanaPacketType) -> None:
        if self.in_queue is not None:
            await self.in_queue.put(paket)


sardana_ingester: HttpSardanaIngester


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    # Load the ML model
    global sardana_ingester
    sardana_ingester = HttpSardanaIngester()
    run_task = asyncio.create_task(sardana_ingester.run())
    run_task.add_done_callback(done_callback)
    yield
    await cancel_and_wait(run_task)
    await sardana_ingester.close()
    # Clean up the ML models and release the resources


app = FastAPI(lifespan=lifespan)


@app.post("/v1/data_desc")
async def post_data_desc(data: SardanaDataDescription = Body(...)) -> None:
    global sardana_ingester
    await sardana_ingester.receive(data)  # SardanaDataDescription(description=payload))
    logger.info("data_desc received %s", data.__repr__())


@app.post("/v1/record_data")
async def post_record_data(data: SardanaRecordData = Body(...)) -> None:
    global sardana_ingester
    await sardana_ingester.receive(data)
    logger.debug("record_data received %s", data.__repr__())


@app.post("/v1/custom_data")
async def post_custom_data(payload: dict[str, Any] = Body(...)) -> None:
    global sardana_ingester
    logger.debug("ignoring custom_data received %s", payload)


@app.post("/v1/record_end")
async def post_record_end(data: SardanaRecordEnd = Body(...)) -> None:
    global sardana_ingester
    await sardana_ingester.receive(data)
    logging.debug("record_end received %s", data.__repr__())



import json
from typing import AsyncGenerator, Any, Iterator

import zmq

from dranspose.event import StreamData
from dranspose.ingester import IsSoftwareTriggered
from dranspose.ingesters.tcp_positioncap import TcpPcapIngester
from dranspose.protocol import StreamName


class SoftTriggerPcapIngester(TcpPcapIngester):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    async def run_source(
        self, stream: StreamName
    ) -> AsyncGenerator[StreamData | IsSoftwareTriggered, None]:
        if stream == "dummy":
            while True:
                yield IsSoftwareTriggered()
        async for streamdata in super().run_source(stream):
            yield streamdata

    def software_trigger(self) -> Iterator[dict[StreamName, StreamData]]:
        while True:
            data = {"temperature": 315.15}
            self._logger.debug("send data %s", data)
            frame = zmq.Frame(json.dumps(data).encode("utf8"))
            yield {StreamName("dummy"): StreamData(typ="JSON", frames=[frame])}



import logging
import uuid
from collections import defaultdict
from typing import Optional, Iterable

from pydantic import validate_call, field_validator, BaseModel

from dranspose.protocol import (
    WorkAssignment,
    StreamName,
    VirtualWorker,
    WorkerName,
    EventNumber,
    WorkerState,
    VirtualConstraint,
    MappingName,
    WorkerTag,
)


class NotYetAssigned(Exception):
    pass


class StillHasWork(Exception):
    pass


logger = logging.getLogger(__name__)


class Map(BaseModel):
    mapping: dict[StreamName, list[Optional[list[VirtualWorker]]]]

    @field_validator("mapping")
    @classmethod
    def same_no_evs_stream(
        cls, v: dict[StreamName, list[Optional[list[VirtualWorker]]]]
    ) -> dict[StreamName, list[Optional[list[VirtualWorker]]]]:
        if len(set(map(len, v.values()))) > 1:
            raise ValueError("length not equal: ", list(map(len, v.values())))
        return v

    def no_events(self) -> int:
        if len(self.mapping) > 0:
            return len(next(iter(self.mapping.values())))
        else:
            return 0

    def items(self) -> Iterable[tuple[StreamName, list[Optional[list[VirtualWorker]]]]]:
        return self.mapping.items()

    def streams(self) -> set[StreamName]:
        return set(self.mapping.keys())

    def all_tags(self) -> set[WorkerTag]:
        tags = set()
        for li in self.mapping.values():
            for frame in li:
                if frame is not None:
                    for vw in frame:
                        tags.update(vw.tags)
        return tags

    def min_workers(self) -> int:
        minimum = 0
        for i in zip(*self.mapping.values()):
            workers = set()
            for val in i:
                if val is not None:
                    workers.update(
                        [v.constraint for v in val if v.constraint is not None]
                    )
            minimum = max(minimum, len(workers))
        return minimum

    @classmethod
    def from_uniform(cls, streams: Iterable[StreamName], ntriggers: int) -> "Map":
        m: dict[StreamName, list[list[VirtualWorker] | None]] = {
            s: [
                [VirtualWorker(constraint=VirtualConstraint(i))]
                for i in range(ntriggers)
            ]
            for s in streams
        }
        return Map(mapping=m)


class ActiveMap(Map):
    assignments: dict[VirtualConstraint, WorkerName] = {}
    all_assignments: dict[
        tuple[EventNumber, StreamName, int], list[WorkerName]
    ] = defaultdict(list)
    complete_events: int = 0
    start_event: EventNumber

    def assign_next(
        self,
        worker: WorkerState,
        all_workers: list[WorkerState],
        completed: Optional[
            EventNumber
        ] = None,  # last event number just completed by this worker
        horizon: int = 0,
    ) -> list[VirtualWorker]:
        assigned_to: list[
            VirtualWorker
        ] = []  # virtual workers the real worker is assigned to, but used nowhere
        maxassign = EventNumber(self.no_events() + 1)
        still_has_work = False
        if completed is not None:
            for evnint in range(completed + 1, self.complete_events - horizon):
                wa = self.get_event_workers(EventNumber(evnint))
                if worker.name in wa.get_all_workers():
                    still_has_work = True
        if still_has_work:
            raise StillHasWork()  # we will not assign the worker if it is assigned until the horizon
        for evnint in range(self.complete_events, self.no_events()):
            evn = EventNumber(evnint)
            for stream, v in self.mapping.items():  # first fill the alls
                assign = v[evn]
                if (
                    assign is not None and len(assign) > 0
                ):  # frame exists and is not discarded
                    for i, vw in enumerate(assign):
                        if vw.constraint is None:
                            required_tags = vw.tags
                            all_worker_names_with_required_tags = [
                                ws.name
                                for ws in all_workers
                                if required_tags.issubset(ws.tags)
                            ]
                            if worker.name in (
                                set(all_worker_names_with_required_tags)
                                - set(self.all_assignments[(evn, stream, i)])
                            ):
                                # assign worker to the "all" of the current stream as it is not yet in it
                                self.all_assignments[(evn, stream, i)].append(
                                    worker.name
                                )
                                assigned_to.append(vw)
                                maxassign = evn  # we assigned the worker to this event, don't continue into the future

            for stream, v in self.mapping.items():  # now fill a specific
                assign = v[evn]
                if (
                    assign is not None and len(assign) > 0
                ):  # frame exists and is not discarded
                    for vw in assign:
                        if vw.constraint is not None:  # not all
                            if vw.constraint not in self.assignments:
                                if vw.tags.issubset(worker.tags):
                                    self.assignments[vw.constraint] = worker.name
                                    self.update_filled(all_workers)
                                    assigned_to.append(vw)
                                    return assigned_to
            if evn == maxassign:
                self.update_filled(all_workers)
                return assigned_to
        if assigned_to == []:
            self.update_filled(all_workers)
        return assigned_to

    def get_event_workers(self, no: EventNumber) -> WorkAssignment:
        ret: dict[StreamName, set[WorkerName]] = defaultdict(set)
        if no > self.complete_events - 1:
            raise NotYetAssigned()
        for s, v in self.mapping.items():
            assign = v[no]
            if assign is None:
                continue
            if len(assign) == 0:
                ret[s] = set()
            for i, vw in enumerate(assign):
                if vw.constraint is None:  # get from all
                    ret[s].update(self.all_assignments[(no, s, i)])
                else:
                    ret[s].add(self.assignments[vw.constraint])
        return WorkAssignment(
            event_number=no, assignments={s: sorted(v) for s, v in ret.items()}
        )

    def update_filled(self, all_workers: list[WorkerState]) -> None:
        for evnint in range(self.complete_events, self.no_events()):
            evn = EventNumber(evnint)
            for stream, v in self.mapping.items():
                assign = v[evn]
                if assign is None:
                    continue
                for i, vw in enumerate(assign):
                    if vw.constraint is None:  # all worker with tags
                        required_tags = vw.tags
                        all_worker_names_with_required_tags = [
                            ws.name
                            for ws in all_workers
                            if required_tags.issubset(ws.tags)
                        ]

                        if set(all_worker_names_with_required_tags) != set(
                            self.all_assignments[(evn, stream, i)]
                        ):
                            return
                    else:
                        if vw.constraint not in self.assignments:
                            return
            self.complete_events = max(0, evn + 1)

    def print(self) -> None:
        print(" " * 5, end="")
        for name in self.mapping:
            print(name.rjust(20), end="")
        print("")
        for evnint in range(self.no_events()):
            evn = EventNumber(evnint)
            print(str(evn + 1).rjust(5), end="")
            for stream in self.mapping:
                val = self.mapping[stream][evn]
                if val is None:
                    print("None".rjust(20), end="")
                    continue
                elif len(val) == 0:
                    print("discard".rjust(20), end="")
                    continue
                else:
                    s = []
                    for i, vw in enumerate(val):
                        if vw.constraint is None:
                            el = (
                                ",".join(map(str, vw.tags))
                                + ":"
                                + " ".join(self.all_assignments[(evn, stream, i)])
                            )
                        else:
                            el = self.assignments.get(vw.constraint, str(vw.constraint))
                        s.append(el)
                    print((";".join(s)).rjust(20), end="")
            print("")


class MappingSequence:
    """
    As the Mapping for many points quickly become large and allocate large amounts of memory,
    we make use of the repetitive structure.
    The parts are a string keyed dictionary of usual maps and
    the offsets are the superstructure on how the parts are combined
    """

    @validate_call
    def __init__(
        self,
        parts: dict[MappingName, dict[StreamName, list[Optional[list[VirtualWorker]]]]],
        sequence: list[MappingName],
        add_start_end: bool = True,
    ) -> None:
        assert (
            set(sequence) - set(parts.keys()) == set()
        ), "all MappingNames in sequence must exist"
        self.parts = {n: Map(mapping=m) for n, m in parts.items()}
        self.all_streams = set()
        for part in self.parts.values():
            self.all_streams.update(part.streams())
        self.all_tags = set()
        for part in self.parts.values():
            self.all_tags.update(part.all_tags())
        if add_start_end:
            start_end_part = Map(
                mapping={
                    stream: [[VirtualWorker(tags={t}) for t in self.all_tags]]
                    for stream in self.all_streams
                }
            )

            self.parts[MappingName("reserved_start_end")] = start_end_part
            sequence.insert(0, MappingName("reserved_start_end"))
            sequence.append(MappingName("reserved_start_end"))

        self.sequence = sequence
        self.uuid = uuid.uuid4()
        self.complete_events = 0

        self.queued_workers: list[
            tuple[WorkerState, list[WorkerState], EventNumber | None]
        ] = []

        self.current_sequence_index = 0
        first_active = ActiveMap(
            start_event=EventNumber(0),
            mapping=self.parts[self.sequence[self.current_sequence_index]].mapping,
        )
        self.active_maps = [first_active]

    def len(self) -> int:
        return sum((self.parts[seq].no_events() for seq in self.sequence))

    def assign_next(
        self,
        worker: WorkerState,
        all_workers: list[WorkerState],
        completed: Optional[EventNumber] = None,
        horizon: int = 0,
        future: int = 10000,
        fill_from_queue: bool = True,
    ) -> list[VirtualWorker]:
        virt = []

        for map in self.active_maps:
            if map.start_event + map.no_events() < self.complete_events:
                # these maps are already complete
                continue
            try:
                virt = map.assign_next(worker, all_workers, completed, horizon)
                if len(virt) > 0:
                    self.complete_events = map.complete_events + map.start_event
                    logger.debug("assigned to %s in map %s", virt, map)
                    return virt
            except StillHasWork:
                return []
        update_completed = True
        while len(virt) == 0 and len(self.active_maps) < len(self.sequence):
            last_map = self.active_maps[-1]
            next_map = ActiveMap(
                start_event=EventNumber(last_map.start_event + last_map.no_events()),
                mapping=self.parts[self.sequence[len(self.active_maps)]].mapping,
            )
            logger.info(
                "start new active map %s starting at %d",
                self.sequence[len(self.active_maps)],
                next_map.start_event,
            )
            self.active_maps.append(next_map)
            if fill_from_queue:
                still_not_assigned = []
                for w, aw, c in self.queued_workers:
                    logger.info("trying to assign %s", w)
                    ret = self.assign_next(w, aw, c, horizon, fill_from_queue=False)
                    # ret = next_map.assign_next(w,aw,c,horizon)
                    if len(ret) == 0:
                        logger.info("worker %s is still not assignable", w)
                        # queued worker is still not possible
                        still_not_assigned.append((w, aw, c))
                    else:
                        logger.info(
                            "finally assigned worker %s in map starting at %d",
                            w,
                            next_map.start_event,
                        )
                self.queued_workers = still_not_assigned
            try:
                virt = next_map.assign_next(worker, all_workers, completed, horizon)
            except StillHasWork:
                return []
            if last_map.complete_events == last_map.no_events() and update_completed:
                self.complete_events = next_map.complete_events + next_map.start_event
                update_completed = False
            if next_map.start_event > self.complete_events + future and len(virt) == 0:
                logger.info(
                    "the worker %s is not useful in the next %d events, queuing it",
                    worker,
                    future,
                )
                self.queued_workers.append((worker, all_workers, completed))
                break
        logger.debug("assigned to %s", virt)

        return virt

    def get_event_workers(self, no: EventNumber) -> WorkAssignment:
        am = None
        if no > self.active_maps[-1].start_event + self.active_maps[-1].no_events():
            raise NotYetAssigned()
        for map in reversed(self.active_maps):
            if map.start_event <= no:
                am = map
                break
        if am is None:
            raise NotYetAssigned()
        wa = am.get_event_workers(EventNumber(no - am.start_event))
        wa.event_number = no
        logger.debug("got event workers for %d: %s", no, wa)
        return wa

    def print(self) -> None:
        print("currently active:", len(self.active_maps))
        for am in self.active_maps:
            print("Starting at event", am.start_event)
            am.print()

    def expand(self) -> dict[StreamName, list[Optional[list[VirtualWorker]]]]:
        mapping: dict[StreamName, list[Optional[list[VirtualWorker]]]] = {
            s: [] for s in self.all_streams
        }
        for seq in self.sequence:
            part = self.parts[seq]
            padding_len = 0
            for stream, li in part.items():
                mapping[stream] += li
                padding_len = len(li)
            for stream in self.all_streams - set(part.streams()):
                mapping[stream] += [None] * padding_len
        return mapping

    def min_workers(self) -> int:
        return max([p.min_workers() for p in self.parts.values()])



import pickle
from types import UnionType

import zmq

from dranspose.data.contrast import ContrastPacket
from dranspose.event import StreamData


def parse(data: StreamData) -> UnionType:
    """
    Parses a contrast packet, which returns a dict, depending on the status of contrast

    Arguments:
        data: a frame comming from contrast

    Returns:
        a ContrastPacket containing a.o. the current status
    """
    assert data.typ == "contrast", "wrong stream name"
    assert data.length == 1, "multipart must be 1"
    frame = data.frames[0]
    if isinstance(frame, zmq.Frame):
        val = pickle.loads(frame.bytes)
    else:
        val = pickle.loads(frame)

    return ContrastPacket.validate_python(val)



from typing import Literal

from pydantic import BaseModel, TypeAdapter, ConfigDict


class SardanaDataDescription(BaseModel):
    """
    Example:
        ```python
        SardanaDataDescription(
            type='DataDescription',
            serialno=20338,
            scandir='/data/staff/femtomax/20230413',
            title='burstscan dummy_mot_01 0.0 0.1 2 50',
            column_desc=[{
                'name': 'point_nb',
                'label': '#Pt No',
                'dtype':
                'int64',
                'shape': []
                }, {
                'min_value': 0,
                'max_value': 0.1,
                'instrument': '',
                'name': 'dummy_mot_01',
                'label': 'dummy_mot_01',
                'dtype': 'float64',
                'shape': [],
                'is_reference': True
                }, {
                'name': 'timestamp',
                'label': 'dt',
                'dtype': 'float64',
                'shape': []}
            ],
            ref_moveables=['dummy_mot_01'],
            estimatedtime=-2.6585786096205566,
            total_scan_intervals=2,
            starttime='Mon Apr 17 14:19:35 2023',
            counters=['tango://b-v-femtomax-csdb-0.maxiv.lu.se:10000/expchan/oscc_02_seq_ctrl/1',
            'tango://b-v-femtomax-csdb-0.maxiv.lu.se:10000/expchan/panda_femtopcap_ctrl/4',
            'tango://b-v-femtomax-csdb-0.maxiv.lu.se:10000/expchan/panda_femtopcap_ctrl/5'],
            scanfile=['stream.daq', 'tests_03.h5']
        )
        ```
    """

    model_config = ConfigDict(extra="allow")
    type: Literal["DataDescription"] = "DataDescription"
    serialno: int
    scandir: str
    title: str


class SardanaRecordData(BaseModel):
    """
    Example:
        ```python
        SardanaRecordData(
            type='RecordData',
            timestamp=0.6104741096496582,
            point_nb=8,
            dummy_mot_01=0
        )
        ```
    """

    model_config = ConfigDict(extra="allow")
    type: Literal["RecordData"] = "RecordData"
    timestamp: float


class SardanaRecordEnd(BaseModel):
    """
    Example:
        ```python
        SardanaRecordEnd(type='RecordEnd')
        ```
    """

    model_config = ConfigDict(extra="allow")
    type: Literal["RecordEnd"] = "RecordEnd"


SardanaPacketType = SardanaDataDescription | SardanaRecordData | SardanaRecordEnd

SardanaPacket = TypeAdapter(
    SardanaDataDescription | SardanaRecordData | SardanaRecordEnd
)
"""
Union type for Sardana packets
"""



import asyncio
import time
import os
from io import BufferedWriter
from typing import AsyncGenerator, Optional, Awaitable, Any, Iterator
from uuid import UUID

import cbor2
import redis.exceptions as rexceptions
import zmq.asyncio
from cbor2 import CBORTag

from pydantic import UUID4, model_validator

from dranspose.distributed import DistributedService, DistributedSettings
from dranspose.event import StreamData, InternalWorkerMessage, message_encoder
from dranspose.helpers.utils import done_callback, cancel_and_wait
from dranspose.protocol import (
    IngesterState,
    StreamName,
    RedisKeys,
    WorkAssignment,
    IngesterName,
    ZmqUrl,
    ParameterName,
    ConnectedWorker,
    IngesterUpdate,
    DistributedStateEnum,
    WorkAssignmentList,
    WorkerName,
)


class IngesterSettings(DistributedSettings):
    ingester_streams: list[StreamName]
    ingester_name: IngesterName
    ingester_url: ZmqUrl = ZmqUrl("tcp://localhost:10000")
    dump_path: Optional[os.PathLike[Any] | str] = None

    @model_validator(mode="before")
    @classmethod
    def create_default_name(cls, data: Any) -> Any:
        if isinstance(data, dict):
            if "ingester_name" not in data and "ingester_streams" in data:
                if isinstance(data["ingester_streams"], list):
                    data["ingester_name"] = (
                        "_".join(map(str, data["ingester_streams"])) + "-ingester"
                    )
        return data


class IsSoftwareTriggered(Exception):
    pass


class Ingester(DistributedService):
    """
    The Ingester class provides a basis to write custom ingesters for any protocol.
    It handles all the forwarding to workers and managing assignments.
    For the instance to do anything, await `run()`
    """

    def __init__(self, settings: Optional[IngesterSettings] = None):
        if settings is None:
            settings = IngesterSettings()
        self._ingester_settings = settings
        state = IngesterState(
            name=self._ingester_settings.ingester_name,
            url=self._ingester_settings.ingester_url,
            streams=self._ingester_settings.ingester_streams,
        )

        super().__init__(state=state, settings=self._ingester_settings)
        self._logger.info(
            "created ingester with state %s and settings %s",
            state,
            self._ingester_settings,
        )
        self.state: IngesterState
        self.active_streams: list[StreamName] = []

        self.dump_file: Optional[BufferedWriter] = None

    def open_socket(self) -> None:
        self.ctx = zmq.asyncio.Context()
        self.out_socket = self.ctx.socket(zmq.ROUTER)
        self.out_socket.setsockopt(zmq.ROUTER_MANDATORY, 1)
        self.out_socket.setsockopt(zmq.TCP_KEEPALIVE, 1)
        self.out_socket.setsockopt(zmq.TCP_KEEPALIVE_IDLE, 300)
        self.out_socket.setsockopt(zmq.TCP_KEEPALIVE_INTVL, 300)
        self.out_socket.bind(f"tcp://*:{self._ingester_settings.ingester_url.port}")

    async def run(self) -> None:
        """
        Main function orchestrating the dependent tasks. This needs to be called from an async context once an instance is created.
        """
        try:
            self.open_socket()
        except Exception as e:
            self._logger.error(
                "unable to open inward facing router socket %s", e.__repr__()
            )

        self.accept_task = asyncio.create_task(self.accept_workers())
        self.accept_task.add_done_callback(done_callback)
        self.work_task = asyncio.create_task(self.work())
        self.work_task.add_done_callback(done_callback)
        self.assign_task = asyncio.create_task(self.manage_assignments())
        self.assign_task.add_done_callback(done_callback)
        self.assignment_queue: asyncio.Queue[WorkAssignment] = asyncio.Queue()
        self.metrics_task = asyncio.create_task(self.update_metrics())
        self.metrics_task.add_done_callback(done_callback)
        self._logger.info("all subtasks running")
        await self.register()

    async def restart_work(
        self, new_uuid: UUID4, active_streams: list[StreamName]
    ) -> None:
        """
        Restarts all work related tasks to make sure no old state is present in a new scan.

        Arguments:
            new_uuid: The uuid of the new mapping
        """
        await cancel_and_wait(self.work_task)
        await cancel_and_wait(self.assign_task)
        self.state.mapping_uuid = new_uuid
        self.active_streams = list(set(active_streams).intersection(self.state.streams))
        self.assignment_queue = asyncio.Queue()
        self.work_task = asyncio.create_task(self.work())
        self.work_task.add_done_callback(done_callback)
        self.assign_task = asyncio.create_task(self.manage_assignments())
        self.assign_task.add_done_callback(done_callback)

    async def finish_work(self) -> None:
        """
        This hook is called when all events of a trigger map were ingested. It is useful for e.g. closing open files.
        """
        self._logger.info("finishing work")
        if self.dump_file:
            self.dump_file.close()
            self._logger.info("closed dump file at finish %s", self.dump_file)
            self.dump_file = None

        await self.redis.xadd(
            RedisKeys.ready(self.state.mapping_uuid),
            {
                "data": IngesterUpdate(
                    state=DistributedStateEnum.FINISHED,
                    ingester=self.state.name,
                ).model_dump_json()
            },
        )

    async def manage_assignments(self) -> None:
        """
        A background task reading assignments from the controller, filering them for the relevant ones and enqueueing them for when the frame arrives.
        """
        self._logger.info("started ingester manage assign task")
        lastev = 0
        while True:
            sub = RedisKeys.assigned(self.state.mapping_uuid)
            try:
                assignments = await self.redis.xread({sub: lastev}, block=1000)
            except rexceptions.ConnectionError:
                break
            if sub not in assignments:
                continue
            assignment_evs = assignments[sub][0]
            self._logger.debug("got assignments %s", assignment_evs)
            for assignment in assignment_evs:
                was = WorkAssignmentList.validate_json(assignment[1]["data"])
                for wa in was:
                    mywa = wa.get_workers_for_streams(self.active_streams)
                    if len(mywa.assignments) > 0:
                        await self.assignment_queue.put(mywa)
                lastev = assignment[0]

    async def _send_workermessages(
        self, workermessages: dict[WorkerName, InternalWorkerMessage]
    ) -> None:
        for worker, message in workermessages.items():
            self._logger.debug(
                "header is %s",
                message.model_dump_json(exclude={"streams": {"__all__": "frames"}}),
            )
            await self.out_socket.send_multipart(
                [worker.encode("ascii")]
                + [
                    message.model_dump_json(
                        exclude={"streams": {"__all__": "frames"}}
                    ).encode("utf8")
                ]
                + message.get_all_frames()
            )
            self._logger.debug("sent message to worker %s", worker)

    async def _get_zmqparts(
        self,
        work_assignment: WorkAssignment,
        sourcegens: dict[
            StreamName, AsyncGenerator[StreamData | IsSoftwareTriggered, None]
        ],
        swtriggen: Iterator[dict[StreamName, StreamData]] | None,
    ) -> dict[StreamName, StreamData]:
        zmqyields: list[Awaitable[StreamData | IsSoftwareTriggered]] = []
        streams: list[StreamName] = []
        for stream in work_assignment.assignments:
            zmqyields.append(anext(sourcegens[stream]))
            streams.append(stream)
        try:
            zmqstreams: list[StreamData | IsSoftwareTriggered] = await asyncio.gather(
                *zmqyields
            )
        except StopAsyncIteration:
            self._logger.warning("stream source stopped before end")
            raise asyncio.exceptions.CancelledError()
        zmqparts: dict[StreamName, StreamData | IsSoftwareTriggered] = {
            stream: zmqpart for stream, zmqpart in zip(streams, zmqstreams)
        }
        self._logger.debug("stream triggered zmqparts %s", zmqparts)
        if swtriggen is not None:
            swparts = next(swtriggen)
            zmqparts.update(swparts)
            # that has to overwrite all IsSoftwareTriggered instance

        return zmqparts

    async def work(self) -> None:
        """
        The heavy liftig function of an ingester. It consumes a generator `run_source()` which
        should be implemented for a specific protocol.
        It then assembles all streams for this ingester and forwards them to the assigned workers.

        Optionally the worker dumps the internal messages to disk. This is useful to develop workers with actual data captured.
        """
        self._logger.info("started ingester work task")
        sourcegens = {stream: self.run_source(stream) for stream in self.active_streams}
        if len(sourcegens) == 0:
            self._logger.warning("this ingester has no active streams, stopping worker")
            return
        swtriggen: Iterator[dict[StreamName, StreamData]] | None = None
        if hasattr(self, "software_trigger"):
            swtriggen = self.software_trigger()
        self.dump_file = None
        self.dump_filename = self._ingester_settings.dump_path
        if self.dump_filename is None and "dump_prefix" in self.parameters:
            val = self.parameters[ParameterName("dump_prefix")].data.decode("utf8")
            if len(val) > 0:
                self.dump_filename = f"{val}{self._ingester_settings.ingester_name}-{self.state.mapping_uuid}.cbors"
        if self.dump_filename:
            self.dump_file = open(self.dump_filename, "ab")
            self._logger.info(
                "dump file %s opened at %s",
                self.dump_filename,
                self.dump_file,
            )
        try:
            took = []
            empties = []
            while True:
                empty = False
                if self.assignment_queue.empty():
                    empty = True
                start = time.perf_counter()
                work_assignment: WorkAssignment = await self.assignment_queue.get()
                if empty:
                    empties.append(work_assignment.event_number)

                zmqparts = await self._get_zmqparts(
                    work_assignment, sourcegens, swtriggen
                )

                if self.dump_file:
                    self._logger.debug("writing dump to path %s", self.dump_filename)
                    allstr = InternalWorkerMessage(
                        event_number=work_assignment.event_number,
                        streams={k: v.get_bytes() for k, v in zmqparts.items()},
                    )
                    try:
                        cbor2.dump(
                            CBORTag(55799, allstr),
                            self.dump_file,
                            default=message_encoder,
                        )
                    except Exception as e:
                        self._logger.error("cound not dump %s", e.__repr__())
                    self._logger.debug("written dump")

                workermessages: dict[WorkerName, InternalWorkerMessage] = {}
                for stream, workers in work_assignment.assignments.items():
                    for worker in workers:
                        if worker not in workermessages:
                            workermessages[worker] = InternalWorkerMessage(
                                event_number=work_assignment.event_number
                            )
                        workermessages[worker].streams[stream] = zmqparts[stream]
                self._logger.debug("workermessages %s", workermessages)
                await self._send_workermessages(workermessages)
                end = time.perf_counter()
                took.append(end - start)
                if len(took) > 1000:
                    self._logger.info(
                        "forwarding took avg %lf, min %f max %f",
                        sum(took) / len(took),
                        min(took),
                        max(took),
                    )
                    self._logger.info("waiting for queue %d of 1000", len(empties))
                    took = []
                    empties = []
                self.state.processed_events += 1
        except asyncio.exceptions.CancelledError:
            self._logger.info("stopping worker")
            for stream in self.active_streams:
                await self.stop_source(stream)
            if self.dump_file:
                self._logger.info(
                    "closing dump file %s at cancelled work", self.dump_file
                )
                self.dump_file.close()
                self.dump_file = None

    async def run_source(
        self, stream: StreamName
    ) -> AsyncGenerator[StreamData | IsSoftwareTriggered, None]:
        """
        This generator must be implemented by the customised subclass. It should return exactly one `StreamData` object
        for every frame arriving from upstream.

        Arguments:
            stream: optionally it received a stream name for which is should yield frames.

        Returns:
            Yield a StreamData object for every received frame.
        """
        yield StreamData(typ="", frames=[])
        return

    async def stop_source(self, stream: StreamName) -> None:
        pass

    async def accept_workers(self) -> None:
        """
        To allow zmq to learn the names of attached workers, they periodically send empty packets.
        There is no information flow directly from workers to ingesters, so we discard the data.
        """
        poller = zmq.asyncio.Poller()
        poller.register(self.out_socket, zmq.POLLIN)
        while True:
            socks = dict(await poller.poll(timeout=1))
            # clean up old workers
            now = time.time()
            self.state.connected_workers = {
                uuid: cw
                for uuid, cw in self.state.connected_workers.items()
                if now - cw.last_seen < 4
            }
            for sock in socks:
                data = await sock.recv_multipart()
                connected_worker = ConnectedWorker(
                    name=data[0], service_uuid=UUID(bytes=data[1])
                )
                fast_publish = False
                if connected_worker.service_uuid not in self.state.connected_workers:
                    fast_publish = True
                self.state.connected_workers[
                    connected_worker.service_uuid
                ] = connected_worker
                self._logger.debug("worker pinnged %s", connected_worker)
                if fast_publish:
                    self._logger.debug("fast publish")
                    await self.publish_config()

    async def close(self) -> None:
        """
        Clean up any open connections
        """
        await cancel_and_wait(self.accept_task)
        await cancel_and_wait(self.work_task)
        await cancel_and_wait(self.metrics_task)
        await cancel_and_wait(self.assign_task)
        await self.redis.delete(RedisKeys.config("ingester", self.state.name))
        await super().close()
        self.ctx.destroy(linger=0)
        self._logger.info("closed ingester")



import asyncio
import logging
from typing import Awaitable, Callable, Any, Coroutine, Optional

import aiohttp

import pytest
import zmq.asyncio
import zmq
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
    EnsembleState,
)

from dranspose.worker import Worker
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_map(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
    stream_orca: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await create_worker(WorkerName("w2"))
    await create_worker(WorkerName("w3"))
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("orca")],
                upstream_url=Url("tcp://localhost:9998"),
                ingester_url=Url("tcp://localhost:10011"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("eiger"), StreamName("orca")})
    async with aiohttp.ClientSession() as session:
        ntrig = 100
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
                # no frames to orca
            },
        )
        assert resp.status == 200

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1))

    content = await wait_for_finish()
    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5000/api/v1/config")
        conf = EnsembleState.model_validate(await st.json())
        msg = []

        for k in conf.workers + conf.ingesters + [conf.reducer]:
            if k is None:
                continue
            msg.append(f"{k.name}:{k.processed_events} -- {k.event_rate}")
            if k.name == "eiger-ingester":
                assert k.processed_events > 0
            elif k.name == "orca-ingester":
                assert k.processed_events == 0

        logging.info("state is \n%s", "\n".join(msg))

    context.destroy()

    print(content)



import asyncio
import json
import logging
import threading
from typing import Any

import h5pyd
import pytest

from dranspose.replay import replay


@pytest.mark.asyncio
async def test_replay(
    tmp_path: Any,
) -> None:
    par_file = tmp_path / "parameters.json"

    with open(par_file, "w") as f:
        json.dump([{"name": "roi1", "data": "[10,20]"}], f)

    stop_event = threading.Event()
    done_event = threading.Event()

    thread = threading.Thread(
        target=replay,
        args=(
            "examples.dummy.worker:FluorescenceWorker",
            "examples.dummy.reducer:FluorescenceReducer",
            None,
            "examples.dummy.source:FluorescenceSource",
            par_file,
        ),
        kwargs={"port": 5010, "stop_event": stop_event, "done_event": done_event},
    )
    thread.start()

    done_event.wait()

    def work() -> None:
        f = h5pyd.File("http://localhost:5010/", "r")
        logging.info("file %s", list(f.keys()))
        logging.info("map %s", list(f["map"].keys()))
        assert list(f.keys()) == ["map"]

    loop = asyncio.get_event_loop()
    await loop.run_in_executor(None, work)

    stop_event.set()

    thread.join()
